/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <gtk/gtk.h>


#include "intl.h"
#include "x_alloc.h"
#include "logging.h"
#include "ui.h"
#include "gtk-utils.h"

#include "gtk-foldersview.h"

enum
  {
    FV_FOLDER_COL,
    FV_NEW_COL,
    FV_UNREAD_COL,
    FV_TOTAL_COL, 
    FV_WEIGHT_COL,
    FV_FOLDERSVIEWNODE_COL,
    FV_N_COLUMNS
  };

struct XePGtkFoldersView_s
{
  XePUIFoldersView ui;
  GtkWidget *widget;
  GtkWidget *optmenu;
  GtkTreeStore *store;
  GtkWidget *tree;
  XePFoldersViewsList *foldersviewslist;
  XePFoldersView *currentfoldersview;
  XePGtkFoldersViewActivateCb activate_cb;
  gpointer activate_data;
  XePGtkFoldersViewButtonCb button_release_cb;
  gpointer button_release_data;
};

typedef struct XePGtkFolderSelectionData_s XePGtkFolderSelectionData;
struct XePGtkFolderSelectionData_s
{
  XePGtkFoldersView *fview;
  XePFolderSelectionCb user_cb;
  void *user_data;
};

static void
GtkFoldersview_destroy_cb(GObject *object, 
			  gpointer afoldersview);

static void
GtkFoldersview_set_view_cb(GtkMenuItem *menuitem, 
			   gpointer data);

static void
GtkFoldersview_set_view_real(XePGtkFoldersView *fview,
			     XePFoldersView *foldersview);

static void
GtkFoldersview_append_node(cntree *tree,
			   GtkTreeStore *store,
			   GtkTreeIter *parent);

static void
GtkFoldersview_set_node(GtkTreeStore *store,
			GtkTreeIter  *iter,
			XePFoldersViewNode *fnode);

static gboolean
GtkFoldersview_select_node(XePGtkFoldersView *fview, 
			   XePFoldersViewNode *fnode,
			   GtkTreeIter *treeiter);

static gboolean
GtkFoldersview_button_release_cb(GtkWidget *widget,
				 GdkEventButton *event,
				 gpointer afview);

static void
GtkFoldersview_row_activated_cb(GtkTreeView *tree,
				GtkTreePath *path,
				GtkTreeViewColumn *column,
				gpointer afview);

static gboolean
GtkFoldersview_refresh_node(GtkTreeModel *model,
			    GtkTreePath  *path,
			    GtkTreeIter  *iter,
			    gpointer      data);
static void
GtkFoldersview_folder_selection_destroy_cb(GObject *object, 
					   gpointer adata);

static gboolean
GtkFoldersview_folder_selection_response_cb(void *adata, 
					    int response);

/* */
XePGtkFoldersView *
a_GtkFoldersview_new(GtkSelectionMode mode)
{
  XePGtkFoldersView *fview;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *label;
  GtkWidget *optmenu;
  GtkWidget *scrollw;
  GtkTreeStore *store;
  GtkWidget *tree;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;
  GtkTreeSelection *selection;

  fview =  x_alloc(1, XePGtkFoldersView);
  if (fview == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }
  fview->ui.type = XETPAN_UI_GTK;

  a_UI_foldersview_init(&fview->ui);

  vbox = gtk_vbox_new(FALSE, 0);

  fview->widget = vbox;

  /* Combox box containing the list of the different views */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
 
  label = gtk_label_new(_("View: "));
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 8);
  gtk_widget_show(label);

  optmenu = gtk_option_menu_new();
  gtk_box_pack_start(GTK_BOX(hbox), optmenu, TRUE, TRUE, 1);
		   
  gtk_widget_show(optmenu);

  fview->optmenu = optmenu;

  gtk_widget_show(hbox);

  /* Tree widget that contain the virtual tree */

  /* First, the data structure */

  store = gtk_tree_store_new(FV_N_COLUMNS, 
			     G_TYPE_STRING, /**< Folder column       */
			     G_TYPE_STRING, /**< New mail count      */
			     G_TYPE_STRING, /**< Unread mail count   */
			     G_TYPE_STRING, /**< Total mail count    */
			     G_TYPE_INT,    /**< Weight              */
			     G_TYPE_POINTER /**< Pointer to the node */
			     );

  fview->store = store;

  /* The real widget */

  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_box_pack_start(GTK_BOX(vbox), scrollw, TRUE, TRUE, 0);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollw),
				      GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw), 
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);
  gtk_widget_show(scrollw);

  tree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
  gtk_container_add(GTK_CONTAINER(scrollw), tree);
  gtk_tree_view_set_hadjustment(GTK_TREE_VIEW(tree),
				gtk_scrolled_window_get_hadjustment
				(GTK_SCROLLED_WINDOW(scrollw)));
  gtk_tree_view_set_vadjustment(GTK_TREE_VIEW(tree),
				gtk_scrolled_window_get_vadjustment
				(GTK_SCROLLED_WINDOW(scrollw)));

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
  gtk_tree_selection_set_mode(selection, mode);
  fview->tree = tree;

  /* Folder name column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Folder"),
						    renderer,
						    "text",
						    FV_FOLDER_COL,
						    "weight",
						    FV_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
  /* New mail count column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("New"),
						    renderer,
						    "text",
						    FV_NEW_COL,
						    "weight",
						    FV_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

  /* Unread mail count column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Unread"),
						    renderer,
						    "text",
						    FV_UNREAD_COL,
						    "weight",
						    FV_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

  /* Total mail count column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Total"),
						    renderer,
						    "text",
						    FV_TOTAL_COL,
						    "weight",
						    FV_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

  gtk_tree_view_set_headers_visible  (GTK_TREE_VIEW(tree), TRUE );
  gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW(tree), FALSE);
  /*  gtk_tree_view_set_reorderable      (GTK_TREE_VIEW(tree), TRUE); */
  gtk_tree_view_set_search_column    (GTK_TREE_VIEW(tree), FV_FOLDER_COL);

  gtk_widget_show(tree);

  /* Complete XePGtkFoldersView structure */
  fview->foldersviewslist = NULL;
  fview->currentfoldersview = NULL;
  fview->activate_cb = NULL;
  fview->activate_data = NULL;
  fview->button_release_cb = NULL;
  fview->button_release_data = NULL;

  /* Signals */
  g_signal_connect(G_OBJECT(fview->widget), "destroy",
		   G_CALLBACK(GtkFoldersview_destroy_cb), fview);
  g_signal_connect_after(G_OBJECT(fview->tree), "button-release-event",
			 G_CALLBACK(GtkFoldersview_button_release_cb), fview);
  g_signal_connect(G_OBJECT(fview->tree), "row-activated",
		   G_CALLBACK(GtkFoldersview_row_activated_cb), fview);
  return fview;
}

static void
GtkFoldersview_destroy_cb(GObject *object, gpointer afview)
{
  a_UI_foldersview_unregister((XePUIFoldersView *) afview);
  free(afview);
}

void
a_GtkFoldersview_free(XePGtkFoldersView *fview)
{
  if (!fview)
    return;

  gtk_widget_destroy(fview->widget);
}

GtkWidget *
a_GtkFoldersview_widget(XePGtkFoldersView *fview)
{
  return fview->widget;
}

void
a_GtkFoldersview_set_activate_callback(XePGtkFoldersView *fview,
				       XePGtkFoldersViewActivateCb user_cb,
				       gpointer user_data)
{
  fview->activate_cb = user_cb;
  fview->activate_data = user_data;
}

void
a_GtkFoldersview_set_button_release_callback(XePGtkFoldersView *fview,
					     XePGtkFoldersViewButtonCb user_cb,
					     gpointer user_data)
{
  fview->button_release_cb = user_cb;
  fview->button_release_data = user_data;

}

void
a_GtkFoldersview_set_list(XePGtkFoldersView *fview,
			  XePFoldersViewsList *foldersviewslist)
{
  clistiter *iter;
  GtkWidget *menu;
  GtkWidget *menuitem;
  XePFoldersView *foldersview;

  if (foldersviewslist)
    {
      menu = gtk_menu_new();
      fview->foldersviewslist = foldersviewslist;
      for (iter = clist_begin(foldersviewslist->list); iter; iter = iter->next)
	{
	  foldersview = (XePFoldersView *) iter->data;
	  menuitem = gtk_menu_item_new_with_label(foldersview->id);
	  gtk_widget_show(menuitem);
	  g_signal_connect(G_OBJECT(menuitem),
			   "activate",
			   G_CALLBACK(GtkFoldersview_set_view_cb),
			   fview);

	  gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
	}

      /* Add the autogenerated view of all mailboxes */

      foldersview = foldersviewslist->mailboxes;
      menuitem = gtk_menu_item_new_with_label(foldersview->id);
      gtk_widget_show(menuitem);
      g_signal_connect(G_OBJECT(menuitem),
		       "activate",
		       G_CALLBACK(GtkFoldersview_set_view_cb),
		       fview);
      
      gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
      
      gtk_widget_show(menu);
      gtk_option_menu_remove_menu(GTK_OPTION_MENU(fview->optmenu));
      gtk_option_menu_set_menu(GTK_OPTION_MENU(fview->optmenu), menu);
      gtk_option_menu_set_history(GTK_OPTION_MENU(fview->optmenu), 0);
      a_GtkFoldersview_set_view(fview, NULL);
    }
  
}

static void
GtkFoldersview_set_view_cb(GtkMenuItem *menuitem, gpointer data)
{
  XePGtkFoldersView *fview = (XePGtkFoldersView *) data;
  GtkWidget *label;

  label = gtk_bin_get_child(GTK_BIN(menuitem));

  a_GtkFoldersview_set_view(fview, gtk_label_get_text(GTK_LABEL(label)));
}

void
a_GtkFoldersview_set_view(XePGtkFoldersView *fview,
			  const char *foldersview_id)
{
  XePFoldersView *foldersview;
  
  A_LOGGING_DEBUG("Showing virtual tree %s", (foldersview_id ? foldersview_id :
					      "(NULL)"));

  foldersview = a_Foldersview_list_get_foldersview(fview->foldersviewslist, 
						   foldersview_id);
  GtkFoldersview_set_view_real(fview, foldersview);
}

static void
GtkFoldersview_set_view_real(XePGtkFoldersView *fview,
			     XePFoldersView *foldersview)
{
  cntree *tree;
  GtkTreeStore *store = fview->store;

  if (foldersview == NULL)
    return;

  tree = foldersview->tree;

  gtk_tree_store_clear(store);
  GtkFoldersview_append_node(tree, store, NULL);
  gtk_tree_view_expand_all(GTK_TREE_VIEW(fview->tree));

  fview->currentfoldersview = foldersview;

  if (GtkFoldersview_select_node(fview, foldersview->lastviewed, NULL))
    {
      if (fview->activate_cb)
	fview->activate_cb(fview, fview->activate_data);
    }
}

static void
GtkFoldersview_append_node(cntree *tree, 
			   GtkTreeStore *store, 
			   GtkTreeIter *parent)
{
  GtkTreeIter *treeiter = NULL;
  cntree *child;
  clistiter *iter;
  XePFoldersViewNode *fnode;

  if (tree == NULL)
    return;

  fnode = (XePFoldersViewNode *) tree->data;
  if (fnode)
    {
      treeiter = x_alloc(1, GtkTreeIter);
      gtk_tree_store_append(store, treeiter, parent);
      GtkFoldersview_set_node(store, treeiter, fnode);
    }
  if (tree->children)
    {
      for (iter = clist_begin(tree->children); iter; iter = iter->next)
	{
	  child = (cntree *) iter->data;
	  GtkFoldersview_append_node(child, store, treeiter);
	}
    }
  x_free(treeiter);
}

static void
GtkFoldersview_set_node(GtkTreeStore *store, 
			GtkTreeIter  *iter,
			XePFoldersViewNode *fnode)
{
  char total[16]; /* log10(2^32) == 10, rounded up to the upper 8 boundary */
  char unseen[16];
  char recent[16];
  XePFolder *folder;
  /* Suppose folder name is not going to be larger than this */
  char fname[XEP_FOLDER_NAME_MAX_LEN]; 
  int weight;

  folder = fnode->folder_ref.folder;

  sprintf(total,  "%d", folder->total);
  sprintf(unseen, "%d", folder->unseen);
  sprintf(recent, "%d", folder->recent);

  weight = ((folder->unseen || folder->recent) ?
	    PANGO_WEIGHT_BOLD :
	    PANGO_WEIGHT_NORMAL);

  if (folder->recent)
    {
      g_snprintf(fname, sizeof(fname), "%s (%d)",
		 fnode->name ? fnode->name : a_Folder_get_id(folder),
		 folder->recent);
    }
  else
    {
      g_strlcpy(fname, fnode->name ? fnode->name : a_Folder_get_id(folder),
		sizeof(fname));
    }
 
  gtk_tree_store_set(store, iter, 
		     FV_FOLDER_COL, fname,
		     FV_NEW_COL, recent,
		     FV_UNREAD_COL, unseen,
		     FV_TOTAL_COL, total,
		     FV_WEIGHT_COL, weight,
		     FV_FOLDERSVIEWNODE_COL, fnode,
		     -1);
}

static gboolean
GtkFoldersview_select_node(XePGtkFoldersView *fview, 
			   XePFoldersViewNode *fnode,
			   GtkTreeIter *treeiter)
{
  gboolean valid;
  GtkTreeIter child;
  GtkTreeModel *treemodel;
  GtkTreeView *treeview = GTK_TREE_VIEW(fview->tree);
  
  treemodel = gtk_tree_view_get_model(treeview);

  valid = gtk_tree_model_iter_children(treemodel, &child, treeiter);
  while (valid)
    {
      XePFoldersViewNode *afoldersviewnode;

      gtk_tree_model_get(treemodel, &child, 
			 FV_FOLDERSVIEWNODE_COL, &afoldersviewnode, -1);
      if (afoldersviewnode == fnode)
	{
	  GtkTreeSelection *selection;
	  selection = gtk_tree_view_get_selection(treeview);
	  gtk_tree_selection_select_iter(selection, &child);
	  valid = FALSE;
	  return TRUE;
	}
      else 
	{
	  if (gtk_tree_model_iter_has_child(treemodel, &child))
	    GtkFoldersview_select_node(fview, fnode, &child);
	  valid = gtk_tree_model_iter_next(treemodel, &child);
	}
    }
  return FALSE;
}

static gboolean
GtkFoldersview_button_release_cb(GtkWidget *widget,
				 GdkEventButton *event,
				 gpointer afview)
{
  XePGtkFoldersView *fview = (XePGtkFoldersView *) afview;
  
  if (fview->button_release_cb)
    return fview->button_release_cb(fview, event, fview->button_release_data);

  return FALSE;
}

static void
GtkFoldersview_row_activated_cb(GtkTreeView *tree,
				GtkTreePath *path,
				GtkTreeViewColumn *column,
				gpointer afview)
{
  XePGtkFoldersView *fview = (XePGtkFoldersView *) afview;
  if (fview->activate_cb)
    fview->activate_cb(fview, fview->activate_data);
}

XePFolderRef *
a_GtkFoldersview_get_selected(XePGtkFoldersView *fview)
{
  GtkTreeView *treeview;
  GtkTreeSelection *selection;
  GtkTreeIter treeiter;
  GtkTreeModel *treemodel;
  XePFoldersViewNode *fnode;
      
  treeview = GTK_TREE_VIEW(fview->tree);
  selection = gtk_tree_view_get_selection(treeview);
  if (gtk_tree_selection_get_selected(selection, &treemodel, &treeiter))
    {
      gtk_tree_model_get(treemodel, &treeiter, 
			 FV_FOLDERSVIEWNODE_COL, &fnode, -1);
      return &(fnode->folder_ref);
    }
  return NULL;
}

carray *
a_GtkFoldersview_get_selected_rows(XePGtkFoldersView *fview)
{
  GtkTreeView *treeview;
  GtkTreeSelection *selection;
  GtkTreeIter treeiter;
  GtkTreeModel *treemodel;
  XePFoldersViewNode *fnode;
  GList *list;
      
  A_LOGGING_TRACE;

  treeview = GTK_TREE_VIEW(fview->tree);
  selection = gtk_tree_view_get_selection(treeview);

  list = gtk_tree_selection_get_selected_rows(selection, &treemodel);

  if (list)
    {
      carray *folder_refs;
      GList *iter;
      folder_refs = carray_new(g_list_length(list));
      for (iter = list; iter; iter = iter->next)
	{
	  GtkTreePath *path = (GtkTreePath *) iter->data;
	  
	  gtk_tree_model_get_iter(treemodel, &treeiter, path);
	  gtk_tree_model_get(treemodel, &treeiter, 
			     FV_FOLDERSVIEWNODE_COL, &fnode, -1);
	  carray_add(folder_refs, &fnode->folder_ref, NULL);
	  gtk_tree_path_free(path);
	}
      g_list_free(list);
      return folder_refs;
    }
  return NULL;
}

void
a_GtkFoldersview_list_refresh(XePGtkFoldersView *fview)
{
  GtkTreeModel *treemodel;
  GtkTreeView *treeview = GTK_TREE_VIEW(fview->tree);

  treemodel = gtk_tree_view_get_model(treeview);
  
  gtk_tree_model_foreach(treemodel,
			 GtkFoldersview_refresh_node,
			 fview);
}

static gboolean
GtkFoldersview_refresh_node(GtkTreeModel *model,
			    GtkTreePath  *path,
			    GtkTreeIter  *iter,
			    gpointer      data)
{
  XePFoldersViewNode *fnode;

  gtk_tree_model_get(model, iter, FV_FOLDERSVIEWNODE_COL, &fnode, -1);
  
  GtkFoldersview_set_node(GTK_TREE_STORE(model), iter, fnode);
  return FALSE;
}

void
a_GtkFoldersview_list_updated(XePGtkFoldersView *fview)
{
  a_GtkFoldersview_set_list(fview, fview->foldersviewslist);
}

/* A Folder Selection Dialog */

GtkWidget *
a_GtkFoldersview_folder_selection(GtkWidget *toplevel, 
				  const gchar *title,
				  const gchar *help,
				  XePFoldersViewsList *foldersviewslist,
				  XePFolderSelectionCb user_cb,
				  void *user_data)
{
  XePGtkFolderSelectionData *data;
  
  data = x_alloc(1, XePGtkFolderSelectionData);
  if (data == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }

  data->fview = a_GtkFoldersview_new(GTK_SELECTION_MULTIPLE);
  a_GtkFoldersview_set_list(data->fview, foldersviewslist);
  data->user_cb = user_cb;
  data->user_data = user_data;

  g_signal_connect(G_OBJECT(XEP_GTK_FOLDERSVIEW_WIDGET(data->fview)),
		   "destroy",
		   G_CALLBACK(GtkFoldersview_folder_selection_destroy_cb),
		   data);

  return a_GtkUtils_dialog_new(_("Folder Selection"),
			       toplevel,
			       title,
			       help,
			       XEP_GTK_FOLDERSVIEW_WIDGET(data->fview),
			       XEP_GTKUTILS_OK | XEP_GTKUTILS_CANCEL,
			       FALSE, FALSE,
			       GtkFoldersview_folder_selection_response_cb,
			       data);
}

static void
GtkFoldersview_folder_selection_destroy_cb(GObject *object, gpointer adata)
{
  XePGtkFolderSelectionData *data = (XePGtkFolderSelectionData *) adata;
  free(data);  
}

static gboolean
GtkFoldersview_folder_selection_response_cb(void *adata, int response)
{
  XePGtkFolderSelectionData *data = (XePGtkFolderSelectionData *) adata;
  XePGtkFoldersView *fview = data->fview;
  XePFolderSelectionCb user_cb = data->user_cb;
  void *user_data = data->user_data;
  carray *folder_refs = NULL;
  
  folder_refs = a_GtkFoldersview_get_selected_rows(fview);
  
  return user_cb(user_data, response, folder_refs);
}


#endif
