/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <gtk/gtk.h>

#include "intl.h"

#include "logging.h"
#include "x_alloc.h"
#include "gtk-utils.h"
#include "gtk-server-editor.h"

#include "gtk-mta-editor.h"

struct XePGtkMTAEditor_s
{
  GtkWidget *widget;
  GtkWidget *name;
  GtkWidget *smtp;
  GtkWidget *shell;
  GtkWidget *notebook;
  XePGtkServerEditor *seditor;
  GtkWidget *command;
  GtkWidget *pipe;
  GtkWidget *arg;

  XePMTA *mta;
  gboolean new_mta;

  XePGtkMTAEditorCb ok_func;
  gpointer ok_data;

  XePGtkMTAEditorCb cancel_func;
  gpointer cancel_data;

};

static gboolean
GtkMTAEditor_delete_event_cb(GtkWidget *widget, 
			     GdkEvent *event, 
			     gpointer amtaeditor);
static void
GtkMTAEditor_destroy_cb(GObject *object, gpointer amtaeditor);

static void
GtkMTAEditor_type_toggled_cb(GtkToggleButton *button, gpointer amtaeditor);

static void
GtkMTAEditor_ok_cb(GtkButton *button, gpointer amtaeditor);

static void
GtkMTAEditor_cancel_cb(GtkButton *button, gpointer amtaeditor);

static void
GtkMTAEditor_update(XePGtkMTAEditor *mtaeditor);

static void
GtkMTAEditor_apply_changes(XePGtkMTAEditor *mtaeditor);

XePGtkMTAEditor *
a_GtkMTAEditor_new(GtkWidget *toplevel)
{
  XePGtkMTAEditor *mtaeditor;
  GtkWidget *vbox;
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *bbox;

  mtaeditor = x_alloc(1, XePGtkMTAEditor);
  if (mtaeditor == NULL)
    {
      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }

  mtaeditor->widget = 
    a_GtkUtils_transient_new(toplevel, _("XetPan Sending Agent Editor"), 
			     -1, -1);
  /* gtk_window_set_modal(GTK_WINDOW(mtaeditor->widget), TRUE);*/

  vbox = gtk_vbox_new(FALSE, 12);
  
  table = gtk_table_new(2, 3, FALSE); 
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);

  mtaeditor->name = a_GtkUtils_table_add_entry_full(GTK_TABLE(table),
						    _("_Name:"),
						    NULL, 
						    _("A unique identifier for this"
						      "sending agent"),
						    0, 2, 0, 1);
  label = gtk_label_new(_("Type:"));
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_widget_show(label);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, GTK_FILL, GTK_SHRINK,
		   0, 0);
  
  mtaeditor->smtp = gtk_radio_button_new_with_mnemonic(NULL,
						       _("_SMTP Server"));
  gtk_widget_show(mtaeditor->smtp);
  gtk_table_attach(GTK_TABLE(table), mtaeditor->smtp, 1, 2, 1, 2, 
		   GTK_FILL, GTK_SHRINK, 0, 0);
  
  mtaeditor->shell = 
    gtk_radio_button_new_with_mnemonic_from_widget(GTK_RADIO_BUTTON(mtaeditor->smtp),
						   _("_External command"));
  gtk_widget_show(mtaeditor->shell);
  gtk_table_attach(GTK_TABLE(table), mtaeditor->shell, 2, 3, 1, 2, 
		   GTK_FILL, GTK_SHRINK, 0, 0);

  gtk_widget_show(table);
  gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);

  mtaeditor->notebook = gtk_notebook_new();
  gtk_notebook_set_show_tabs(GTK_NOTEBOOK(mtaeditor->notebook), FALSE);
  gtk_notebook_set_show_border(GTK_NOTEBOOK(mtaeditor->notebook), FALSE);

  /* Server Editor Page */
  mtaeditor->seditor = a_GtkServerEditor_new();
  gtk_widget_show(XEP_SEDITOR_WIDGET(mtaeditor->seditor));

  gtk_notebook_append_page(GTK_NOTEBOOK(mtaeditor->notebook), 
			   XEP_SEDITOR_WIDGET(mtaeditor->seditor),
			   NULL);

  /* Shell Command Page */
  table = gtk_table_new(4, 3, FALSE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);

  label = a_GtkUtils_label_new_bold(_("Shell Command:"));
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_widget_show(label);
  gtk_table_attach(GTK_TABLE(table), label, 0, 3, 0, 1, GTK_FILL, GTK_SHRINK,
		   0, 0);

  mtaeditor->command = a_GtkUtils_table_add_entry(GTK_TABLE(table),
						  _("_Command:"), NULL, 
						  _("Shell command to execute"),
						  1, 1);
  mtaeditor->pipe = gtk_radio_button_new_with_mnemonic(NULL, 
						       _("_Pipe message to "
							 "command's standard "
							 "input"));
  gtk_widget_show(mtaeditor->pipe);
  gtk_table_attach(GTK_TABLE(table), mtaeditor->pipe, 1, 3, 3, 4,
		   GTK_FILL, GTK_SHRINK, 0, 0);

  mtaeditor->arg = 
    gtk_radio_button_new_with_mnemonic_from_widget(GTK_RADIO_BUTTON(mtaeditor->pipe), 
						   _("_Append message to "
						     "command's arguments"));
  
  gtk_widget_show(mtaeditor->arg);
  gtk_table_attach(GTK_TABLE(table), mtaeditor->arg, 1, 3, 2, 3,
		   GTK_FILL, GTK_SHRINK, 0, 0);

  gtk_widget_show(table);
  gtk_notebook_append_page(GTK_NOTEBOOK(mtaeditor->notebook), table, NULL);

  gtk_box_pack_start(GTK_BOX(vbox), mtaeditor->notebook, TRUE, TRUE, 0);
  gtk_widget_show(mtaeditor->notebook);

  bbox = gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 12);

  button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  gtk_widget_show(button);
  gtk_container_add(GTK_CONTAINER(bbox), button);
  g_signal_connect(G_OBJECT(button), "clicked",
		   G_CALLBACK(GtkMTAEditor_cancel_cb), mtaeditor);

  button = gtk_button_new_from_stock(GTK_STOCK_OK);
  gtk_widget_show(button);
  gtk_container_add(GTK_CONTAINER(bbox), button);
  g_signal_connect(G_OBJECT(button), "clicked",
		   G_CALLBACK(GtkMTAEditor_ok_cb), mtaeditor);

  gtk_widget_show(bbox);
  gtk_box_pack_start(GTK_BOX(vbox), bbox, FALSE, FALSE, 0);

  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(mtaeditor->widget), vbox);

  /* Signals */
  g_signal_connect(G_OBJECT(mtaeditor->widget), "destroy",
		   G_CALLBACK(GtkMTAEditor_destroy_cb), mtaeditor);

  g_signal_connect(G_OBJECT(mtaeditor->widget), "delete-event",
		   G_CALLBACK(GtkMTAEditor_delete_event_cb), mtaeditor);

  g_signal_connect(G_OBJECT(mtaeditor->smtp), "toggled",
		   G_CALLBACK(GtkMTAEditor_type_toggled_cb),
		   mtaeditor);

  mtaeditor->mta = NULL;
  mtaeditor->new_mta = FALSE;
  mtaeditor->ok_func = NULL;
  mtaeditor->ok_data = NULL;
  mtaeditor->cancel_func = NULL;
  mtaeditor->cancel_data = NULL;

  return mtaeditor;
}

static void
GtkMTAEditor_destroy_cb(GObject *object, gpointer amtaeditor)
{
  XePGtkMTAEditor *mtaeditor = (XePGtkMTAEditor *) amtaeditor;

  a_GtkMTAEditor_free(mtaeditor);
}

gboolean
GtkMTAEditor_delete_event_cb(GtkWidget *widget, 
			     GdkEvent *event, 
			     gpointer amtaeditor)
{
  XePGtkMTAEditor *mtaeditor = (XePGtkMTAEditor *) amtaeditor;
  
  a_GtkMTAEditor_hide(mtaeditor);
  return TRUE;
}

void
a_GtkMTAEditor_free(XePGtkMTAEditor *mtaeditor)
{
  a_GtkServerEditor_free(mtaeditor->seditor);
  x_free(mtaeditor);
}

static void
GtkMTAEditor_type_toggled_cb(GtkToggleButton *button, gpointer amtaeditor)
{
  XePGtkMTAEditor *mtaeditor = (XePGtkMTAEditor *) amtaeditor;
  XePMTAType type;

  type = (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mtaeditor->smtp)) ?
	  XEP_MTA_SMTP :
	  XEP_MTA_COMMAND);

  a_MTA_set_type(mtaeditor->mta, type);

  GtkMTAEditor_update(mtaeditor);
}

void
a_GtkMTAEditor_show(XePGtkMTAEditor *mtaeditor)
{
  gtk_window_present(GTK_WINDOW(mtaeditor->widget));
}

void
a_GtkMTAEditor_hide(XePGtkMTAEditor *mtaeditor)
{
  gtk_widget_hide(mtaeditor->widget);

  if (mtaeditor->new_mta)
    {
      a_MTA_free(mtaeditor->mta);
      mtaeditor->mta = NULL;
      mtaeditor->new_mta = FALSE;
    }

  if (mtaeditor->cancel_func)
    mtaeditor->cancel_func(mtaeditor, mtaeditor->ok_data);
}

static void
GtkMTAEditor_cancel_cb(GtkButton *button, gpointer amtaeditor)
{
  XePGtkMTAEditor *mtaeditor = (XePGtkMTAEditor *) amtaeditor;

  a_GtkMTAEditor_hide(mtaeditor);
}

static void
GtkMTAEditor_ok_cb(GtkButton *button, gpointer amtaeditor)
{
  XePGtkMTAEditor *mtaeditor = (XePGtkMTAEditor *) amtaeditor;
  
  gtk_widget_hide(mtaeditor->widget);

  GtkMTAEditor_apply_changes(mtaeditor);

  if (mtaeditor->ok_func)
    mtaeditor->ok_func(mtaeditor, mtaeditor->ok_data);

  if (mtaeditor->new_mta)
    {
      a_MTA_free(mtaeditor->mta);
      mtaeditor->mta = NULL;
      mtaeditor->new_mta = FALSE;
    }
}

void
a_GtkMTAEditor_set_mta(XePGtkMTAEditor *mtaeditor, XePMTA *mta)
{
  if (mtaeditor->new_mta)
    {
      a_MTA_free(mtaeditor->mta);
      mtaeditor->new_mta = FALSE;
    }

  if (mta == NULL)
    {
      mta = a_MTA_new();
      mtaeditor->new_mta = TRUE;
      a_MTA_set_type(mta, XEP_MTA_SMTP);
    }

  mtaeditor->mta = mta;

  GtkMTAEditor_update(mtaeditor);
}

XePMTA *
a_GtkMTAEditor_get_mta(XePGtkMTAEditor *mtaeditor)
{
  XePMTA *mta = mtaeditor->mta;

  mtaeditor->mta = NULL;
  mtaeditor->new_mta = FALSE;

  return mta;
}

void
a_GtkMTAEditor_set_ok_callback(XePGtkMTAEditor *mtaeditor,
			       XePGtkMTAEditorCb func,
			       gpointer data)
{
  mtaeditor->ok_func = func;
  mtaeditor->ok_data = data;
}

void
a_GtkMTAEditor_set_cancel_callback(XePGtkMTAEditor *mtaeditor,
				   XePGtkMTAEditorCb func,
				   gpointer data)
{
  mtaeditor->cancel_func = func;
  mtaeditor->cancel_data = data;
}

static void
GtkMTAEditor_update(XePGtkMTAEditor *mtaeditor)
{
  XePMTA *mta = mtaeditor->mta;
  const char *id;
  const char *cmd;
  gboolean pipe = FALSE;
  XePMTAType type;

  if (mta == NULL)
    return;


  id = a_MTA_get_id(mta);
  gtk_entry_set_text(GTK_ENTRY(mtaeditor->name), id ? id : "");

  type = a_MTA_get_type(mta);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(type == XEP_MTA_SMTP ?
						 mtaeditor->smtp :
						 mtaeditor->shell),
			       TRUE);

  switch (type)
    {
    case XEP_MTA_COMMAND:
      cmd = a_MTA_get_command(mta);
      if (cmd && cmd[0] == '|')
	{
	  cmd++;
	  pipe = TRUE;
	}

      gtk_entry_set_text(GTK_ENTRY(mtaeditor->command), cmd ? cmd : "");
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pipe ? 
						     mtaeditor->pipe :
						     mtaeditor->arg), TRUE);
      gtk_notebook_set_current_page(GTK_NOTEBOOK(mtaeditor->notebook), 1);
      break;
    case XEP_MTA_SMTP:
      a_GtkServerEditor_set_server(mtaeditor->seditor, a_MTA_get_server(mta));
      a_GtkServerEditor_set_type(mtaeditor->seditor, XEP_SERVEREDITOR_SMTP);
      gtk_notebook_set_current_page(GTK_NOTEBOOK(mtaeditor->notebook), 0);
      break;
    default:
      A_LOGGING_ERROR("Attempt to setting an MTA with an unknown type");
    }
}

static void
GtkMTAEditor_apply_changes(XePGtkMTAEditor *mtaeditor)
{
  XePMTAType type;
  XePMTA *mta;
  char *cmd;
  gboolean pipe;

  mta = mtaeditor->mta;
  type = a_MTA_get_type(mta);

  a_MTA_set_id(mta, gtk_entry_get_text(GTK_ENTRY(mtaeditor->name)));

  switch (type)
    {
    case XEP_MTA_SMTP:
      a_GtkServerEditor_apply_changes(mtaeditor->seditor);
      break;
    case XEP_MTA_COMMAND:
      pipe = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mtaeditor->pipe));
      cmd = g_strdup_printf((pipe ? "|%s" : "%s"), 
			    gtk_entry_get_text(GTK_ENTRY(mtaeditor->command)));
      a_MTA_set_command(mta, cmd);
      g_free(cmd);
      break;
    default:
      A_LOGGING_ERROR("Should not be reached: applying MTA changes with "
		      "unknown MTA type");
    }      
}
#endif
