/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h> /* close() */
#include <gtk/gtk.h>

#include "intl.h"

#include "logging.h"
#include "x_alloc.h"

#include <libetpan/charconv.h>

#include "defines.h"
#include "gtk-utils.h"
#include "utils.h"

/**! WARNING: These must be in sync with XePMessageHeaderType!!! */
static const char *gtkutils_header_name_tags [] = 
  {
    "date",
    "from",
    "sender",
    "reply-to",
    "to",
    "cc",
    "bcc",
    "message-id",
    "in-reply-to",
    "references",
    "subject",
    "comments",
    "keywords",
    "other"
  };

/**! WARNING: These must be in sync with XePMessageHeaderType!!! */
static const char *gtkutils_header_value_tags [] = 
  {
    "date-value",
    "from-value",
    "sender-value",
    "reply-to-value",
    "to-value",
    "cc-value",
    "bcc-value",
    "message-id-value",
    "in-reply-to-value",
    "references-value",
    "subject-value",
    "comments-value",
    "keywords-value",
    "other-value"
  };

static GtkTooltips *tips;

typedef struct XePGtkUtilsDialog_s XePGtkUtilsDialog;

struct XePGtkUtilsDialog_s
{
  GtkWidget *toplevel;
  XePGtkUtilsDialogCb user_cb;
  void *user_data;
  gboolean kill_on_close;
};

typedef struct XePGtkUtilsDialogString_s XePGtkUtilsDialogString;

struct XePGtkUtilsDialogString_s
{
  GtkWidget *toplevel;
  XePGtkUtilsDialogStringCb callback;
  void *data;
  GtkWidget *entry;
};

typedef struct XePGtkUtilsTreeButton_s XePGtkUtilsTreeButton;

struct XePGtkUtilsTreeButton_s
{
  GtkWidget *button;
  XePGtkUtilsTreeCb callback;
  void *data;
};
    
struct XePGtkUtilsTree_s
{
  GtkWidget *widget;
  GtkWidget *buttons;
  GtkWidget *treeview;
  GtkTreeStore *treestore;
  clist *list;
  cntree *tree;
  const char *description;
  int type;
  int weight;
  void *user_data1;
  void *user_data2;
  int is_expanded;
  char *path;
  XePGtkUtilsTreeButton add;
  XePGtkUtilsTreeButton remove;
  XePGtkUtilsTreeButton edit;
  XePGtkUtilsTreeButton top;
  XePGtkUtilsTreeButton changed;  /* Actually there is no button, but
				   * as it can store callbacks, we use it
				   */
  XePGtkUtilsTreeButton refresh;  /* Actually there is no button, but
				   * as it can store callbacks, we use it
				   */
  XePGtkUtilsTreeButton expanded; /* Actually there is no button, but
				   * as it can store callbacks, we use it
				   */
};

typedef struct XePGtkUtilsTreeItem_s XePGtkUtilsTreeItem;

struct XePGtkUtilsTreeItem_s
{
  char *id;
};

/* Static functions declarations */

static void
GtkUtils_dialog_destroy_cb(GObject *object, gpointer data);

static void
GtkUtils_dialog_response_cb(GtkDialog *dialog,
			    gint response,
			    gpointer data);

static void
GtkUtils_dialog_get_string_destroy_cb(GObject *object, gpointer data);

static void
GtkUtils_dialog_get_string_response_cb(GtkDialog *dialog,
				       gint response,
				       gpointer data);

static void
GtkUtils_dialog_open_filesel_cb(GtkButton *button, gpointer data);

static int
GtkUtils_dialog_filesel_response_cb(void *adialog_str, const char *filename);

static void
GtkUtils_dialog_alert_response_cb(GtkDialog *dialog,
				  gint response,
				  gpointer data);
static void
GtkUtils_text_buffer_append_image_data_cb(GdkPixbufLoader *loader,
					  gpointer data);

static gboolean
GtkUtils_paned_button_release_cb(GtkWidget *widget,
				 GdkEventButton *event,
				 gpointer data);
static void
GtkUtils_file_selection_ok_cb(GtkWidget *widget,
			      gpointer data);

static void
GtkUtils_dir_selection_ok_cb(GtkWidget *widget,
			     gpointer data);

static void
GtkUtils_tree_destroy_cb(GObject *awidget, gpointer autils_tree);

static void
GtkUtils_tree_append(GtkTreeStore *treestore, 
		     GtkTreeIter *parent, 
		     cntree *ntree);

static void
GtkUtils_tree_set_callback(XePGtkUtilsTreeButton *button,
			   XePGtkUtilsTreeCb callback,
			   void *data);

static int
GtkUtils_tree_get_selected(XePGtkUtilsTree *utils_tree, 
			   GtkTreePath **path,
			   const char **description,
			   int *type,
			   int *weight,
			   void **user_data1,
			   void **user_data2);

static void 
GtkUtils_tree_add_cb(GtkButton *button, gpointer data);

static void 
GtkUtils_tree_remove_cb(GtkButton *button, gpointer data);

static void 
GtkUtils_tree_edit_cb(GtkButton *button, gpointer data);

static void 
GtkUtils_tree_top_cb(GtkButton *button, gpointer data);

static void 
GtkUtils_tree_collapsed_cb(GtkTreeView *treeview,GtkTreeIter *expanded_iter,
			   GtkTreePath *expanded_path, gpointer data);

static void 
GtkUtils_tree_expanded_cb(GtkTreeView *treeview,GtkTreeIter *expanded_iter,
			  GtkTreePath *expanded_path, gpointer data);
/* Functions definitions */

GtkWidget *
a_GtkUtils_dialog_new(const char *title,
		      GtkWidget *parent,
		      const char *label_text,
		      const char *help_text,
		      GtkWidget *child,
		      int buttons,
		      gboolean modal,
		      gboolean kill_on_close,
		      XePGtkUtilsDialogCb user_cb,
		      void *user_data)
{
  XePGtkUtilsDialog *custom_dialog;
  GtkWidget *dialog;
  GtkWidget *vbox;
  GtkWidget *label;

  custom_dialog = x_alloc(1, XePGtkUtilsDialog);
  if (custom_dialog == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }

  dialog = gtk_dialog_new_with_buttons(title, GTK_WINDOW(parent), 
				       modal ? GTK_DIALOG_MODAL : 0, NULL);
  gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);

  vbox = gtk_vbox_new(FALSE, 12);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
		     vbox,
		     TRUE, TRUE, 0);

  if (label_text)
    {
      label = a_GtkUtils_label_new_bold(label_text);
      gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
    }

  if (help_text)
    {
      label = a_GtkUtils_label_new_italic(help_text);
      gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
      gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
    }

  gtk_box_pack_start(GTK_BOX(vbox), child, TRUE, TRUE, 0);

  if (buttons & XEP_GTKUTILS_CANCEL)
    {
      gtk_dialog_add_button(GTK_DIALOG(dialog), 
			    GTK_STOCK_CANCEL,
			    GTK_RESPONSE_CANCEL);
      gtk_dialog_set_default_response (GTK_DIALOG(dialog), 
				       GTK_RESPONSE_CANCEL);
    }

  if (buttons & XEP_GTKUTILS_OK)
    {
      gtk_dialog_add_button(GTK_DIALOG(dialog), 
			    GTK_STOCK_OK, 
			    GTK_RESPONSE_OK);
      gtk_dialog_set_default_response (GTK_DIALOG(dialog), 
				       GTK_RESPONSE_OK);
    }

  custom_dialog->toplevel = dialog;
  custom_dialog->user_cb = user_cb;
  custom_dialog->user_data = user_data;
  custom_dialog->kill_on_close = kill_on_close;

  g_signal_connect(G_OBJECT(dialog),
		   "response",
		   G_CALLBACK(GtkUtils_dialog_response_cb),
		   custom_dialog);

  /* GtkDialog sends the cancel response and destroys
   * the widget. We avoid it.
   */
  g_signal_connect_swapped(G_OBJECT(dialog), "delete-event",
			   G_CALLBACK(gtk_true), NULL);

  g_signal_connect(G_OBJECT(dialog), "destroy",
		   G_CALLBACK(GtkUtils_dialog_destroy_cb),
		   custom_dialog);

  gtk_widget_show_all(dialog);
  return dialog;
}

static void
GtkUtils_dialog_destroy_cb(GObject *object, gpointer data)
{
  XePGtkUtilsDialog *custom_dialog = (XePGtkUtilsDialog *) data;

  free(custom_dialog);
}

static void
GtkUtils_dialog_response_cb(GtkDialog *dialog,
			    gint response,
			    gpointer data)
{
  XePGtkUtilsDialog *custom_dialog = (XePGtkUtilsDialog *) data;
  XePGtkUtilsDialogCb user_cb;
  void *user_data;

  user_cb  = custom_dialog->user_cb;
  user_data = custom_dialog->user_data;

  /* user_cb must return TRUE to notify that operation
   * succeded and dialog can be closed
   */
  if (user_cb(user_data, response))
    {
      if (custom_dialog->kill_on_close)
	{
	  gtk_widget_destroy(GTK_WIDGET(dialog));
	}
      else
	{
	  gtk_widget_hide(GTK_WIDGET(dialog));
	}
    }
}

void
a_GtkUtils_dialog_get_string(const char *title, 
			     GtkWidget *parent,
			     const char *label_text,
			     const char *help_text,
			     const char *default_str,
			     int buttons,
			     gboolean modal,
			     XePGtkUtilsDialogStringCb callback,
			     void *data)
{
  GtkWidget *dialog;
  GtkWidget *entry;
  GtkWidget *label;
  GtkWidget *hbox;
  GtkWidget *vbox;

  XePGtkUtilsDialogString *dialog_str;

  dialog_str = x_alloc(1, XePGtkUtilsDialogString);
  if (dialog_str == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return;
    }

  dialog = gtk_dialog_new_with_buttons(title, GTK_WINDOW(parent), 
				       modal ? GTK_DIALOG_MODAL : 0, NULL);
  gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);

  vbox = gtk_vbox_new(FALSE, 12);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
		     vbox,
		     TRUE, TRUE, 0);

  label = a_GtkUtils_label_new_bold_large(title);
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

  hbox = gtk_hbox_new(FALSE, 12);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  if (label_text)
    {
      label = gtk_label_new(label_text);
      gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    }

  entry = gtk_entry_new();
  if (default_str)
    gtk_entry_set_text(GTK_ENTRY(entry), default_str);
  gtk_editable_select_region(GTK_EDITABLE(entry), 0, -1);
  gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  if (buttons & XEP_GTKUTILS_FILESEL)
    {
      GtkWidget *button;
      button = gtk_button_new_from_stock(GTK_STOCK_OPEN);
      gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
      g_signal_connect(G_OBJECT(button), "clicked",
		       G_CALLBACK(GtkUtils_dialog_open_filesel_cb),
		       dialog_str);
		       
    }
  if (help_text)
    {
      label = a_GtkUtils_label_new_italic(help_text);
      gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
      gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
    }

  if (buttons & XEP_GTKUTILS_CANCEL)
    {
      gtk_dialog_add_button(GTK_DIALOG(dialog), 
			    GTK_STOCK_CANCEL,
			    GTK_RESPONSE_CANCEL);
      gtk_dialog_set_default_response (GTK_DIALOG(dialog), 
				       GTK_RESPONSE_CANCEL);
    }

  if (buttons & XEP_GTKUTILS_OK)
    {
      gtk_dialog_add_button(GTK_DIALOG(dialog), 
			    GTK_STOCK_OK, 
			    GTK_RESPONSE_OK);
      gtk_dialog_set_default_response (GTK_DIALOG(dialog), 
				       GTK_RESPONSE_OK);
    }

  dialog_str->toplevel = dialog;
  dialog_str->callback = callback;
  dialog_str->data     = data;
  dialog_str->entry    = entry;

  g_signal_connect(G_OBJECT(dialog),
		   "response",
		   G_CALLBACK(GtkUtils_dialog_get_string_response_cb),
		   dialog_str);

  /* GtkDialog sends the cancel response and destroys
   * the widget. We avoid it.
   */
  g_signal_connect_swapped(G_OBJECT(dialog), "delete-event",
			   G_CALLBACK(gtk_true), NULL);

  g_signal_connect(G_OBJECT(dialog),
		   "destroy",
		   G_CALLBACK(GtkUtils_dialog_get_string_destroy_cb),
		   dialog_str);

  gtk_widget_show_all(dialog);
}

static void
GtkUtils_dialog_get_string_destroy_cb(GObject *object, gpointer data)
{
  XePGtkUtilsDialogString *dialog_str = (XePGtkUtilsDialogString *) data;
  free(dialog_str);
}

static void
GtkUtils_dialog_get_string_response_cb(GtkDialog *dialog,
				      gint response,
				      gpointer data)
{
  XePGtkUtilsDialogString *dialog_str = (XePGtkUtilsDialogString *) data;
  XePGtkUtilsDialogStringCb callback;
  GtkWidget *entry;
  void *user_data;
  gchar *answer = NULL;

  callback  = dialog_str->callback;
  user_data = dialog_str->data;
  entry     = dialog_str->entry;

  if (response == GTK_RESPONSE_OK)
    {
      answer = gtk_editable_get_chars(GTK_EDITABLE(entry), 0, -1);
    }

  gtk_widget_destroy(GTK_WIDGET(dialog));

  callback(user_data, response, answer);
}

static void
GtkUtils_dialog_open_filesel_cb(GtkButton *button, gpointer data)
{
  XePGtkUtilsDialogString *dialog_str = (XePGtkUtilsDialogString *) data;
  
  a_GtkUtils_file_selection(_("Select a file"), dialog_str->toplevel,
			    NULL,
			    GtkUtils_dialog_filesel_response_cb,
			    dialog_str);
}

static int
GtkUtils_dialog_filesel_response_cb(void *adialog_str, const char *filename)
{
  XePGtkUtilsDialogString *dialog_str =
    (XePGtkUtilsDialogString *) adialog_str;

  gtk_entry_set_text(GTK_ENTRY(dialog_str->entry), filename);
  return 0;  
}

int
a_GtkUtils_dialog_alert(const char *title, 
			GtkWidget *parent,
			int type,
			const char *message_text,
			const char *help_text,
			int buttons,
			gboolean modal,
			gboolean async,
			void (*user_callback)(void *adata, int aresponse),
			void *user_data)
{
  GtkWidget *dialog;
  GtkWidget *hbox;
  GtkWidget *image;
  GtkWidget *vbox;
  GtkWidget *label;

  dialog = gtk_dialog_new_with_buttons((type == XETPAN_INFO_MESSAGE ?
					_("XetPan - Info") :
					type == XETPAN_WARNING_MESSAGE ?
					_("XetPan - Warning") :
					_("XetPan - Error")),
				       GTK_WINDOW(parent), 
				       GTK_DIALOG_NO_SEPARATOR |
				       (modal ? GTK_DIALOG_MODAL : 0),
				       NULL);

  hbox = gtk_hbox_new(FALSE, 12);
  gtk_container_set_border_width(GTK_CONTAINER(hbox), 12);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
		     hbox,
		     TRUE, TRUE, 0);
  

  image = gtk_image_new_from_stock((type == XETPAN_INFO_MESSAGE ?
				    GTK_STOCK_DIALOG_INFO :
				    type == XETPAN_WARNING_MESSAGE ?
				    GTK_STOCK_DIALOG_WARNING :
				    GTK_STOCK_DIALOG_ERROR),
				   GTK_ICON_SIZE_DIALOG);
  gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 12);

  vbox = gtk_vbox_new(FALSE, 12);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);

  label = a_GtkUtils_label_new_bold_large(title);
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

  if (message_text)
    {
      label = gtk_label_new(message_text);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
    }

  if (help_text)
    {
      label = a_GtkUtils_label_new_italic(help_text);
      gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
      gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
    }
  
  if (buttons & XEP_GTKUTILS_CANCEL)
    {
      gtk_dialog_add_button(GTK_DIALOG(dialog), 
			    GTK_STOCK_CANCEL,
			    GTK_RESPONSE_CANCEL);
      gtk_dialog_set_default_response (GTK_DIALOG(dialog), 
				       GTK_RESPONSE_CANCEL);
    }

  if (buttons & XEP_GTKUTILS_OK)
    {
      gtk_dialog_add_button(GTK_DIALOG(dialog), 
			    GTK_STOCK_OK, 
			    GTK_RESPONSE_OK);
      gtk_dialog_set_default_response (GTK_DIALOG(dialog), 
				       GTK_RESPONSE_OK);
    }

  gtk_widget_show_all(dialog);

  if (async == FALSE)
    {
      gint response;
      response = gtk_dialog_run(GTK_DIALOG(dialog));

      if (user_callback)
	user_callback(user_data, response);

      gtk_widget_destroy(dialog);

      return response;
    }
  
  g_signal_connect(G_OBJECT(dialog),
		   "response",
		   G_CALLBACK(GtkUtils_dialog_alert_response_cb),
		   dialog);
  g_object_set_data(G_OBJECT(dialog), "user-callback", user_callback);
  g_object_set_data(G_OBJECT(dialog), "user-data", user_data);
  return 0;
}

static void
GtkUtils_dialog_alert_response_cb(GtkDialog *dialog,
				  gint response,
				  gpointer data)
{
  void *user_data;
  int (*user_callback)(void *adata, int aresponse);

  user_callback = g_object_get_data(G_OBJECT(dialog), "user-callback");
  user_data = g_object_get_data(G_OBJECT(dialog), "user-data");
  
  gtk_widget_destroy(GTK_WIDGET(dialog));

  if (user_callback)
    user_callback(user_data, response);
}

GtkWidget*
a_GtkUtils_label_new_bold_large(const char *text)
{
  /* TODO: maybe only one instance for the attributes and let the
   * reference counting do its job?
   */
  PangoAttrList *attr_list;
  PangoAttribute *weight_attr;
  PangoAttribute *scale_attr;

  GtkWidget *label;

  attr_list = pango_attr_list_new();
  weight_attr = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
  weight_attr->start_index = 0;
  weight_attr->end_index = -1;

  scale_attr = pango_attr_scale_new(PANGO_SCALE_LARGE);
  scale_attr->start_index = 0;
  scale_attr->end_index = -1;

  pango_attr_list_insert(attr_list, weight_attr);
  pango_attr_list_insert(attr_list, scale_attr);

  label = gtk_label_new(text);
  gtk_label_set_attributes(GTK_LABEL(label), attr_list);

  pango_attr_list_unref(attr_list);

  return label;
}

GtkWidget*
a_GtkUtils_label_new_bold(const char *text)
{
  /* TODO: maybe only one instance for the attributes and let the
   * reference counting do its job?
   */
  PangoAttrList *attr_list;
  PangoAttribute *weight_attr;

  GtkWidget *label;

  attr_list = pango_attr_list_new();
  weight_attr = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
  weight_attr->start_index = 0;
  weight_attr->end_index = -1;

  pango_attr_list_insert(attr_list, weight_attr);

  label = gtk_label_new(text);
  gtk_label_set_attributes(GTK_LABEL(label), attr_list);

  pango_attr_list_unref(attr_list);

  return label;
}

GtkWidget*
a_GtkUtils_label_new_italic(const char *text)
{
  /* TODO: maybe only one instance for the attributes and let the
   * reference counting do its job?
   */
  PangoAttrList *attr_list;
  PangoAttribute *style_attr;

  GtkWidget *label;

  attr_list = pango_attr_list_new();

  style_attr = pango_attr_style_new(PANGO_STYLE_ITALIC);
  style_attr->start_index = 0;
  style_attr->end_index = -1;

  pango_attr_list_insert(attr_list, style_attr);

  label = gtk_label_new(text);
  gtk_label_set_attributes(GTK_LABEL(label), attr_list);

  pango_attr_list_unref(attr_list);

  return label;
}

GtkWidget *
a_GtkUtils_transient_new(GtkWidget *toplevel,
			 const char *title, 
			 int width,
			 int height)
{
  GtkWidget *window;

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_screen(GTK_WINDOW(window), 
                        gtk_window_get_screen(GTK_WINDOW(toplevel)));
  gtk_window_set_transient_for(GTK_WINDOW(window),
                               GTK_WINDOW(toplevel));
  gtk_window_set_destroy_with_parent(GTK_WINDOW(window), TRUE);
  gtk_window_set_title(GTK_WINDOW(window), title);
  gtk_window_set_default_size(GTK_WINDOW(window), width, height);
  gtk_container_set_border_width(GTK_CONTAINER(window), 12);

  return window;
}

GtkWidget *
a_GtkUtils_toplevel_new(const char *terminal, 
			const char *title, 
			const char **xpm_data,
			int width,
			int height)
{
  GtkWidget *toplevel;
  GdkPixbuf *icon;

  GdkScreen *screen = NULL;
  GdkDisplay *display = NULL;

  toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  if (terminal)
    {
      display = gdk_display_open(terminal);
      if (display)
	{
	  screen = gdk_display_get_default_screen(display);
	  if (screen)
	    {
	      gtk_window_set_screen(GTK_WINDOW(toplevel), screen);
	      A_LOGGING_INFO("Using screen %s", 
			     gdk_screen_make_display_name(screen));
	    }
	  else
	    A_LOGGING_ERROR("Could not get default screen on display %s",
			    terminal);
	}
      else
	A_LOGGING_ERROR("Could not open display %s", terminal);
    }

  gtk_window_set_title(GTK_WINDOW(toplevel), title);

  gtk_window_set_default_size(GTK_WINDOW(toplevel), width, height);

  if (xpm_data)
    {
      icon = gdk_pixbuf_new_from_xpm_data(xpm_data);
      gtk_window_set_icon(GTK_WINDOW(toplevel), icon);
      g_object_unref(G_OBJECT(icon));
    }

  gtk_widget_realize(toplevel);

  return toplevel;
}

void
a_GtkUtils_file_viewer_new(const char *terminal,
			   const char *title,
			   const char **xpm_data,
			   int width,
			   int height,
			   const char *filename,
			   const char *data,
			   size_t size,
			   const char *charset)
{
  GtkWidget *toplevel;
  GtkWidget *textview;
  GtkTextIter iter;
  GtkWidget *scrollw;
  GtkTextBuffer *buffer;
  char *magic_mime_type = NULL;
  const char *magic_charset = NULL;
  int fd;
  struct stat stat_buf;
  char *maddr;

  if (width == 0)
    width = 450;
  if (height == 0)
    height = 500;

  toplevel = a_GtkUtils_toplevel_new(terminal, title, xpm_data, width, height);

  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollw),
				      GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw), 
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(toplevel), scrollw);
  
  textview = a_GtkUtils_text_view_new(NULL, 0, FALSE);
  gtk_container_add(GTK_CONTAINER(scrollw), textview);

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(textview));

  /* Find the charset */
  if (charset == NULL)
    {
      if (filename)
	{
	  magic_mime_type = a_Utils_magic_file_mime_type(filename);
	}
      else if (data) 
	{
	  magic_mime_type = a_Utils_magic_data_mime_type(data, size);
	}
      if ((magic_mime_type) && 
	  (magic_charset = strstr(magic_mime_type, "charset=")))
	{
	  magic_charset += strlen("charset=");
	}
    }
  else 
    {
      magic_charset = charset;
    }

  gtk_text_buffer_get_end_iter(buffer, &iter);

  /* Insert data */
  if (filename)
    {
      fd = open(filename, O_RDONLY);
      if (fd)
	{
	  if (fstat(fd, &stat_buf) == 0)
	    {	  
	      if (stat_buf.st_size)
		{
		  maddr = mmap(NULL, stat_buf.st_size, PROT_READ, MAP_SHARED,
			       fd, 0);
		  a_GtkUtils_text_buffer_convert_insert(buffer, &iter,
							maddr,
							stat_buf.st_size,
							magic_charset);
		  munmap(maddr, stat_buf.st_size);
		}
	      else
		{
		  gtk_text_buffer_set_text(buffer, _("<Empty file>"), -1);
		}
	    }
	  else
	    {
	      A_LOGGING_SYSTEM_ERROR("fstat(): %s", strerror(errno));
	    }
	  close(fd);
	}
      else
	{
	  A_LOGGING_ERROR("Error while opening file '%s': %s",
			  filename, strerror(errno));
	}
    }
  else if (data)
    {
      a_GtkUtils_text_buffer_convert_insert(buffer, &iter,
					    data, size,
					    magic_charset);
    }
  else
    {
      gtk_text_buffer_set_text(buffer, _("<Empty data>"), -1);
    }

  x_free(magic_mime_type);

  g_signal_connect_swapped(G_OBJECT(toplevel), "delete-event", 
		   G_CALLBACK(gtk_widget_destroy), toplevel);

  gtk_widget_show_all(toplevel);
}

static gboolean
GtkUtils_stop_button_release_cb(GtkWidget *widget,
				GdkEventButton *event,
				gpointer data)
{
  return TRUE;
}

void
a_GtkUtils_text_view_stop_button_release(GtkWidget *widget)
{
  g_signal_connect_after(G_OBJECT(widget), "button-release-event",
			 G_CALLBACK(GtkUtils_stop_button_release_cb),
			 NULL);
}

GtkWidget *
a_GtkUtils_text_view_new(const char *text, 
			 gint len, 
			 gboolean editable)
{
  GtkWidget *text_view;
  GtkTextBuffer *buffer;
  GtkTextIter iter;

  text_view = gtk_text_view_new();
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_view));
  gtk_text_view_set_left_margin(GTK_TEXT_VIEW(text_view), 2);
  gtk_text_view_set_right_margin(GTK_TEXT_VIEW(text_view), 2);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view), editable);
  gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(text_view), editable);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(text_view), GTK_WRAP_WORD);
  if (text)
    gtk_text_buffer_set_text(buffer, text, len);

  /* Place cursor at start. If we don't do it, then a click on a button in
   * another tab of the notebook will scroll down the text view
   */
  gtk_text_buffer_get_start_iter(buffer, &iter);
  gtk_text_buffer_place_cursor(buffer, &iter);


  a_GtkUtils_text_view_stop_button_release(text_view);

  return text_view;
}

void
a_GtkUtils_text_buffer_create_tags(GtkTextBuffer *buffer)
{
#if GTK_HAS_INVISIBLE_TAG
  gtk_text_buffer_create_tag(buffer, "invisible",
			     "invisible", TRUE,
			     NULL);
#endif

  gtk_text_buffer_create_tag(buffer, "from",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "from-value",
			     "weight", PANGO_WEIGHT_NORMAL,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "to",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "to-value",
			     "weight", PANGO_WEIGHT_NORMAL,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "subject",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "subject-value",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);
  gtk_text_buffer_create_tag(buffer, "cc",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "cc-value",
			     "weight", PANGO_WEIGHT_NORMAL,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);
  gtk_text_buffer_create_tag(buffer, "bcc",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "bcc-value",
			     "weight", PANGO_WEIGHT_NORMAL,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);
  gtk_text_buffer_create_tag(buffer, "reply-to",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "reply-to-value",
			     "weight", PANGO_WEIGHT_NORMAL,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "date",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "date-value",
			     "weight", PANGO_WEIGHT_NORMAL,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "other",
			     "weight", PANGO_WEIGHT_BOLD,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);

  gtk_text_buffer_create_tag(buffer, "other-value",
			     "weight", PANGO_WEIGHT_NORMAL,
			     "scale", PANGO_SCALE_MEDIUM,
			     "style", PANGO_STYLE_NORMAL,
			     NULL);
  gtk_text_buffer_create_tag(buffer, "reply-selection",
			     "weight", PANGO_WEIGHT_BOLD,
			     NULL);
  gtk_text_buffer_create_tag(buffer, "monospace",
			     "font", "monospace 8",
			     NULL);
}

void
a_GtkUtils_text_buffer_get_header_tag(XePMessageHeaderType type,
				      const char **header_name_tag,
				      const char **header_value_tag)
{
  *header_name_tag = gtkutils_header_name_tags[type];
  *header_value_tag = gtkutils_header_value_tags[type];
}

void
a_GtkUtils_text_buffer_append(GtkTextBuffer *buffer,
			      const char *text,
			      int size)
{
  GtkTextIter iter;
  gtk_text_buffer_get_end_iter(buffer, &iter);
  gtk_text_buffer_insert(buffer, &iter, text, size);
  gtk_text_buffer_place_cursor(buffer, &iter);
}

void
a_GtkUtils_text_buffer_convert_insert(GtkTextBuffer *buffer,
				      GtkTextIter *iter,
				      const char *text,
				      size_t text_len,
				      const char *source_charset)
{
  int r;
  char *converted = NULL;
  size_t converted_len = 0;
  
  r = charconv_buffer(GTK_DEST_CHARSET, 
		      (char *)(source_charset ? source_charset : DEFAULT_SOURCE_CHARSET),
		      (char *)text, text_len,
		      &converted, &converted_len);
  if (r == MAIL_CHARCONV_NO_ERROR)
    {
      gtk_text_buffer_insert(buffer, iter, converted, converted_len);
      charconv_buffer_free(converted);
    }
  else
    {
      gtk_text_buffer_insert(buffer, iter, text, text_len);
    }
}

int
a_GtkUtils_text_buffer_insert_image_file(GtkTextBuffer *buffer,
					 GtkTextIter *iter,
					 const char *filename)
{
  GdkPixbuf *pixbuf;
  GError *error = NULL;

  int r = -1;

  pixbuf = gdk_pixbuf_new_from_file(filename, &error);

  if (error)
    {
      A_LOGGING_ERROR("gdk_pixbuf_new_from_file(): %s", error->message);
      g_error_free(error);
    }
  else
    {
      gtk_text_buffer_insert_pixbuf(buffer, iter, pixbuf);
      g_object_unref(G_OBJECT(pixbuf));
      r = 0;
    }

  return r;
}

static void
GtkUtils_text_buffer_append_image_data_cb(GdkPixbufLoader *loader,
					  gpointer data)
{
  GtkTextBuffer *buffer = (GtkTextBuffer *) data;
  GdkPixbuf *pixbuf;

  pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);

  if (pixbuf)
    {
      GtkTextIter iter;
      gtk_text_buffer_get_end_iter(buffer, &iter);
      gtk_text_buffer_insert_pixbuf(buffer, &iter, pixbuf);
      g_object_unref(G_OBJECT(pixbuf));
    }
}

int
a_GtkUtils_text_buffer_append_image_data(GtkTextBuffer *buffer,
					 const char *data,
					 size_t len)
{
  GdkPixbufLoader *pixbuf_loader;
  GError *error = NULL;
  int r;
  
  r = -1;

  pixbuf_loader = gdk_pixbuf_loader_new();
  
  if (error)
    {
      A_LOGGING_ERROR("gdk_pixbuf_loader_new(): %s", error->message);
      g_error_free(error);
      error = NULL;
      return r;
    }

  g_signal_connect(G_OBJECT(pixbuf_loader), "area_prepared",
		   G_CALLBACK(GtkUtils_text_buffer_append_image_data_cb),
		   buffer);
  
  gdk_pixbuf_loader_write(pixbuf_loader, data, len, &error);
  if (error)
    {
      A_LOGGING_ERROR("gdk_pixbuf_loader_write(): %s", error->message);
      g_error_free(error);
    }
  else
    {
      gdk_pixbuf_loader_close(pixbuf_loader, &error);
      if (error)
	{
	  A_LOGGING_ERROR("gdk_pixbuf_loader_close(): %s", error->message);
	  g_error_free(error);
	}
      else 
	{
	  GdkPixbuf *pixbuf;
	  
	  pixbuf = gdk_pixbuf_loader_get_pixbuf(pixbuf_loader);
	  r = 0;
	}
    }
  return r;
}

int
a_GtkUtils_text_buffer_insert_html(GtkTextBuffer *buffer,
				   GtkTextIter *iter,
				   const char *text,
				   size_t text_len,
				   const char *source_charset)
{
  const char *cmd = "LC_ALL=en_IE.utf8 w3m -dump -T text/html";
  int fd_html[2];
  int fd_out[2];
  int r = 0;
  ssize_t w;
  size_t c;
  GtkTextIter start;
  GtkTextMark *mark;

  /* html renderer input pipe */
  if (pipe(fd_html))
    {
      A_LOGGING_ERROR("pipe(): %s", strerror(errno));
      r = -1;
      return r;
    }
  /* html renderer output pipe */
  if (pipe(fd_out))
    {
      A_LOGGING_ERROR("pipe(): %s", strerror(errno));
      close(fd_html[0]);
      close(fd_html[1]);
      r = -1;
      return r;
    }

  a_Utils_exec_cmd(cmd, fd_html[0], fd_out[1], -1, NULL, NULL, 0);
  close(fd_html[0]);
  close(fd_out[1]);

  /* send html text to renderer */
  c = 0;
  w = 0;
  while (c < text_len && w >= 0)
    {
      w = write(fd_html[1], text + c, text_len - c);
      if (w > 0)
	c += w;
      if (w < 0 && errno == EINTR)
	w = 0;
    }
  if (w < 0)
    {
      A_LOGGING_ERROR("write(): %s", strerror(errno));
      r = -1;
    }
  close(fd_html[1]);

  /* Read renderer output */
  mark = gtk_text_buffer_create_mark(buffer, NULL, iter, TRUE);

  #define HTML_BUF_SIZE 4096
  c = 0;
  do
    {
      char buf[HTML_BUF_SIZE];
      w = read(fd_out[0], buf, sizeof(buf));
      if (w > 0)
	gtk_text_buffer_insert(buffer, iter, buf, w);
    } while(w > 0);

  if (w < 0)
    {
      A_LOGGING_ERROR("read(): %s", strerror(errno));
      r = -1;
    }
  #undef HTML_BUF_SIZE

  close(fd_out[0]);
  gtk_text_buffer_get_iter_at_mark(buffer, &start, mark);
  gtk_text_buffer_apply_tag_by_name(buffer, "monospace", &start, iter);
  return r;
}

void
a_GtkUtils_paned_shrink_control(GtkWidget *paned, gpointer data)
{
  int *position = (int *) data;
  *position = -1;

  g_signal_connect(G_OBJECT(paned), 
		   "button-release-event",
		   G_CALLBACK(GtkUtils_paned_button_release_cb),
		   data);
}

static gboolean
GtkUtils_paned_button_release_cb(GtkWidget *widget,
				 GdkEventButton *event,
				 gpointer data)
{
  GtkPaned *paned = GTK_PANED(widget);
  int *stored_position = (int *) data;
  int actual_position;
  int reference_position;
  int opposite_position;
  int new_position;

  if (event->button != 2 && event->button != 3)
    return FALSE;

  actual_position = gtk_paned_get_position(paned);

  if (event->button == 2)
    {
      reference_position = paned->min_position;
      opposite_position = paned->max_position;
    }
  else
    {
      reference_position = paned->max_position;
      opposite_position = paned->min_position;
    }

  if (actual_position == reference_position)
    {
      new_position = *stored_position;

      if (new_position >= paned->max_position)
	new_position = (paned->max_position + paned->min_position) >> 1;
    }
  else
    {
      if (actual_position != opposite_position)
	*stored_position = actual_position;

      new_position = reference_position;
    }

  gtk_paned_set_position(paned, new_position);

  return TRUE;  
}

gboolean
a_GtkUtils_filename_is_supported_image(const char *filename)
{
  GSList *format_list;
  GSList *format_item;
  GdkPixbufFormat *format;
  gchar **known_extensions;
  int i;
  gboolean found = FALSE;
  const char *extension;

  if (filename == NULL)
    return FALSE;

  /* Check if there is an extension, if not assume the whole filename is an
     extension */
  if ((extension = strrchr(filename, '.')))
    {
      extension++;
      if (*extension == 0x00)
	return FALSE;
    }

  /* TODO: see if making format_list a static variable is really needed 
   * if gdk_pixbuf_get_formats() is slow 
   */
  
  format_list = gdk_pixbuf_get_formats();

  if (format_list == NULL)
    return FALSE;

  for (format_item = format_list; 
       format_item && !found;
       format_item = format_item->next)
    {
      format = (GdkPixbufFormat *) format_item->data;
      known_extensions = gdk_pixbuf_format_get_extensions(format);
      for (i = 0; known_extensions[i] && !found; i++)
	{
	  if (strcasecmp(known_extensions[i], extension) == 0)
	    found = TRUE;
	}
    }

  g_slist_free(format_list);
  return found;
}

void
a_GtkUtils_file_selection(const char *title,
			  GtkWidget *toplevel,
			  const char *filename,
			  int (*callback)(void *adata, const char *fname),
			  void *data)
{
  GtkWidget *file_selection;
  file_selection = gtk_file_selection_new(title);
  
  gtk_window_set_screen(GTK_WINDOW(file_selection),
			gtk_window_get_screen(GTK_WINDOW(toplevel)));

  gtk_window_set_transient_for(GTK_WINDOW(file_selection),
                               GTK_WINDOW(toplevel));

  gtk_window_set_modal(GTK_WINDOW(file_selection), TRUE);
  gtk_window_set_destroy_with_parent(GTK_WINDOW(file_selection), TRUE);

  if (filename)
    {
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selection),
				      filename);
    }

  g_object_set_data(G_OBJECT(file_selection), "user-callback", callback);
  g_object_set_data(G_OBJECT(file_selection), "user-data", data);
  g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(file_selection)->ok_button),
		   "clicked",
		   G_CALLBACK(GtkUtils_file_selection_ok_cb),
		   file_selection);

  g_signal_connect_swapped(G_OBJECT
			   (GTK_FILE_SELECTION(file_selection)->cancel_button),
			   "clicked",
			   G_CALLBACK(gtk_widget_destroy), 
			   (gpointer) file_selection); 
  
  gtk_widget_show(file_selection);
}

static void
GtkUtils_file_selection_ok_cb(GtkWidget *widget,
			      gpointer data)
{
  const char *selected_filename;
  GtkFileSelection *file_selection = (GtkFileSelection *) data;
  void *user_data;
  int (*user_callback)(void *adata, const char *filename);

  user_callback = g_object_get_data(G_OBJECT(file_selection), "user-callback");
  user_data = g_object_get_data(G_OBJECT(file_selection), "user-data");

  selected_filename = gtk_file_selection_get_filename(file_selection);
  if (user_callback(user_data, selected_filename) != 0)
    {
      A_LOGGING_ERROR(_("Error during file operation."));
    }
  else
    {
      gtk_widget_destroy(GTK_WIDGET(data));
    }

}

void
a_GtkUtils_dir_selection(const char *title,
			 GtkWidget *toplevel,
			 const char *dirname,
			 int (*callback)(void *adata, const char *fname),
			 void *data)
{
  GtkWidget *file_selection;
  GtkFileSelection *fs;
  file_selection = gtk_file_selection_new(title);
  fs = GTK_FILE_SELECTION(file_selection);

  gtk_window_set_screen(GTK_WINDOW(file_selection),
			gtk_window_get_screen(GTK_WINDOW(toplevel)));

  gtk_window_set_transient_for(GTK_WINDOW(file_selection),
                               GTK_WINDOW(toplevel));

  gtk_window_set_modal(GTK_WINDOW(file_selection), TRUE);
  gtk_window_set_destroy_with_parent(GTK_WINDOW(file_selection), TRUE);

  if (dirname)
    {
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selection),
				      dirname);
    }

  /* Hide file selection list */
  gtk_widget_hide(fs->file_list->parent);
  gtk_widget_hide(fs->selection_text);
  gtk_widget_hide(fs->selection_entry);
  gtk_widget_hide(fs->fileop_ren_file);
  gtk_widget_hide(fs->fileop_del_file);

  g_object_set_data(G_OBJECT(file_selection), "user-callback", callback);
  g_object_set_data(G_OBJECT(file_selection), "user-data", data);
  g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(file_selection)->ok_button),
		   "clicked",
		   G_CALLBACK(GtkUtils_dir_selection_ok_cb),
		   file_selection);

  g_signal_connect_swapped(G_OBJECT
			   (GTK_FILE_SELECTION(file_selection)->cancel_button),
			   "clicked",
			   G_CALLBACK(gtk_widget_destroy), 
			   (gpointer) file_selection); 
  
  gtk_widget_show(file_selection);
}


static void
GtkUtils_dir_selection_ok_cb(GtkWidget *widget,
			     gpointer data)
{
  GtkTreeModel *model = NULL;
  GtkTreeIter iter;
  GtkTreeSelection *selection;
  const char *selected_item = NULL;
  const char *selected_path = NULL;
  char *complete_path;
  GtkFileSelection *fs = (GtkFileSelection *) data;
  void *user_data;
  int (*user_callback)(void *adata, const char *filename);

  user_callback = g_object_get_data(G_OBJECT(fs), "user-callback");
  user_data = g_object_get_data(G_OBJECT(fs), "user-data");

  selected_path = gtk_file_selection_get_filename(fs);

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(fs->dir_list));
  if (gtk_tree_selection_get_selected(selection, &model, &iter))
    {
      gtk_tree_model_get(model, &iter, 0, &selected_item, -1);
    }
  
  complete_path = g_strdup_printf("%s%s",
				  selected_path,
				  (selected_item ? selected_item : ""));

  if (user_callback && user_callback(user_data, complete_path) != 0)
    {
      A_LOGGING_ERROR(_("Error during file operation."));
    }
  else
    {
      gtk_widget_destroy(GTK_WIDGET(data));
    }
  g_free(complete_path);
}

GtkWidget *
a_GtkUtils_button_new_from_stock(const char *stock_id, 
				 const char *text,
				 gboolean use_markup,
				 GtkIconSize icon_size)
{
  GtkWidget *button;
  GtkWidget *alignment;
  GtkWidget *image;
  GtkWidget *hbox;
  GtkWidget *label;
 
  button = gtk_button_new();

  alignment = gtk_alignment_new(0.5, 0.5, 0, 0);
  gtk_container_add(GTK_CONTAINER(button), alignment);
  gtk_widget_show(alignment);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(alignment), hbox);
  gtk_widget_show(hbox);

  image = gtk_image_new_from_stock(stock_id, icon_size);
  gtk_box_pack_start(GTK_BOX (hbox), image, FALSE, FALSE, 0);
  gtk_widget_show(image);

  if (text)
    {
      label = gtk_label_new_with_mnemonic(text);
      gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);
      gtk_label_set_use_markup(GTK_LABEL(label), use_markup);
      gtk_widget_show_all(label);
    }


  return button;  
}

gchar *
a_GtkUtils_make_display_name(GtkWidget *toplevel)
{
  gchar *display;
  display = gdk_screen_make_display_name(gtk_window_get_screen
				 (GTK_WINDOW(toplevel)));
  return display;
}

XePGtkUtilsTree *
a_GtkUtils_tree_new(GtkOrientation orientation, gboolean reordable,
		    GtkPolicyType hpolicy, GtkPolicyType vpolicy)
{
  XePGtkUtilsTree * utils_tree;
  GtkTreeViewColumn *column;
  GtkCellRenderer *renderer;

  utils_tree = x_alloc(1, XePGtkUtilsTree);

  if (utils_tree == NULL)
    {
      A_LOGGING_ERROR("x_alloc(): %s", strerror(errno));
      return NULL;
    }

  /* Initialize members */

  utils_tree->changed.button   = NULL; /* there is no changed button, */
  utils_tree->changed.callback = NULL; /* it is just a container for  */
  utils_tree->changed.data     = NULL; /* the "changed" callback      */

  utils_tree->refresh.button   = NULL; /* there is no changed button, */
  utils_tree->refresh.callback = NULL; /* it is just a container for  */
  utils_tree->refresh.data     = NULL; /* the "refresh" callback      */

  utils_tree->expanded.button   = NULL; /* there is no changed button, */
  utils_tree->expanded.callback = NULL; /* it is just a container for  */
  utils_tree->expanded.data     = NULL; /* the "expanded" callback     */

  utils_tree->is_expanded      = 0;
  utils_tree->tree             = NULL;
  utils_tree->list             = NULL;
  utils_tree->description      = NULL;
  utils_tree->type             = 0;
  utils_tree->weight          = 0;
  utils_tree->user_data1       = NULL;
  utils_tree->user_data2       = NULL;
  utils_tree->path             = NULL;

  /* List widget */

  utils_tree->widget = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(utils_tree->widget),
				 hpolicy, vpolicy);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(utils_tree->widget),
				      GTK_SHADOW_IN);

  /* Tree Store */

  /* TODO: check if G_TYPE_STRING means the string is copied */
  utils_tree->treestore = gtk_tree_store_new(TREE_COL_NB,
					     G_TYPE_STRING,   /* Description */
					     G_TYPE_INT,      /* User TYPE */
					     G_TYPE_INT,      /* Weight */
					     G_TYPE_POINTER,  /* User data 1 */
					     G_TYPE_POINTER); /* User data 2 */
  
  /* Tree View */

  utils_tree->treeview = 
    gtk_tree_view_new_with_model(GTK_TREE_MODEL(utils_tree->treestore));
  gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(utils_tree->treeview), 
				    FALSE);
  gtk_tree_view_set_reorderable(GTK_TREE_VIEW(utils_tree->treeview), 
				reordable);
  gtk_container_add(GTK_CONTAINER(utils_tree->widget), utils_tree->treeview);
  gtk_widget_show(utils_tree->treeview);
  
  /* The treeview holds a reference and is responsible for freeing
   * the treestore
   */

  g_object_unref(G_OBJECT(utils_tree->treestore));

  /* Description column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(NULL, 
						    renderer,
						    "text",
						    TREE_DESCRIPTION_COL,
						    "weight",
						    TREE_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(utils_tree->treeview), column);

  /* Button box */

  if (orientation == GTK_ORIENTATION_HORIZONTAL)
    utils_tree->buttons = gtk_hbox_new(TRUE, 6);
  else
    utils_tree->buttons = gtk_vbox_new(TRUE, 6);

  utils_tree->edit.button = gtk_button_new_from_stock(GTK_STOCK_PROPERTIES);
  gtk_box_pack_start(GTK_BOX(utils_tree->buttons), utils_tree->edit.button,
		     TRUE, TRUE, 0);

  utils_tree->top.button = gtk_button_new_from_stock(GTK_STOCK_GOTO_TOP);
  gtk_box_pack_start(GTK_BOX(utils_tree->buttons), utils_tree->top.button,
		     TRUE, TRUE, 0);

  utils_tree->add.button = gtk_button_new_from_stock(GTK_STOCK_ADD);
  gtk_box_pack_start(GTK_BOX(utils_tree->buttons), utils_tree->add.button,
		     TRUE, TRUE, 0);

  utils_tree->remove.button = gtk_button_new_from_stock(GTK_STOCK_REMOVE);
  gtk_box_pack_start(GTK_BOX(utils_tree->buttons), utils_tree->remove.button,
		     TRUE, TRUE, 0);

  /* Signal callbacks */

  g_signal_connect(G_OBJECT(utils_tree->widget), "destroy",
		   G_CALLBACK(GtkUtils_tree_destroy_cb), utils_tree);
  g_signal_connect(G_OBJECT(utils_tree->treeview), "row-expanded",
		   G_CALLBACK(GtkUtils_tree_expanded_cb), utils_tree);
  g_signal_connect(G_OBJECT(utils_tree->treeview), "row-collapsed",
		   G_CALLBACK(GtkUtils_tree_collapsed_cb), utils_tree);
  g_signal_connect(G_OBJECT(utils_tree->add.button), "clicked",
		   G_CALLBACK(GtkUtils_tree_add_cb), utils_tree);
  g_signal_connect(G_OBJECT(utils_tree->remove.button), "clicked",
		   G_CALLBACK(GtkUtils_tree_remove_cb), utils_tree);
  g_signal_connect(G_OBJECT(utils_tree->edit.button), "clicked",
		   G_CALLBACK(GtkUtils_tree_edit_cb), utils_tree);
  g_signal_connect(G_OBJECT(utils_tree->top.button), "clicked",
		   G_CALLBACK(GtkUtils_tree_top_cb), utils_tree);


  return utils_tree;
}

void
a_GtkUtils_tree_free(XePGtkUtilsTree *utils_tree)
{
  if (utils_tree == NULL)
    return;

  x_free(utils_tree->path);
  free(utils_tree);
}

static void
GtkUtils_tree_destroy_cb(GObject *awidget, gpointer autils_tree)
{
  XePGtkUtilsTree *utils_tree = (XePGtkUtilsTree *) autils_tree;
  a_GtkUtils_tree_free(utils_tree);
}

GtkWidget *
a_GtkUtils_tree_get_tree_widget(XePGtkUtilsTree *utils_tree)
{
  return utils_tree->widget;
}

GtkWidget *
a_GtkUtils_tree_get_buttons_widget(XePGtkUtilsTree *utils_tree)
{
  return utils_tree->buttons;
}

GtkTreeStore *
a_GtkUtils_tree_get_tree_store(XePGtkUtilsTree *utils_tree)
{
  return utils_tree->treestore;
}

GtkWidget *
a_GtkUtils_tree_get_tree_view(XePGtkUtilsTree *utils_tree)
{
  return utils_tree->treeview;
}

void
a_GtkUtils_tree_set_list(XePGtkUtilsTree *utils_tree,
			 clist *list)
{
  utils_tree->tree = NULL;
  utils_tree->list = list;
}

void
a_GtkUtils_tree_set_tree(XePGtkUtilsTree *utils_tree,
			cntree *ntree)
{
  utils_tree->tree = ntree;
  utils_tree->list = NULL;
}

GtkTreeSelection *
a_GtkUtils_tree_get_selection(XePGtkUtilsTree *utils_tree)
{
  return gtk_tree_view_get_selection(GTK_TREE_VIEW(utils_tree->treeview));
}

void
a_GtkUtils_tree_selection_set_mode(XePGtkUtilsTree *utils_tree, 
				   GtkSelectionMode mode)
{
  GtkTreeSelection *selection;

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(utils_tree->treeview));
  gtk_tree_selection_set_mode(selection, mode);
}

int
a_GtkUtils_tree_selected_is_expanded(XePGtkUtilsTree *utils_tree)
{
  return utils_tree->is_expanded;
}

void
a_GtkUtils_tree_set_path(XePGtkUtilsTree *utils_tree, GtkTreePath *path)
{
  x_free(utils_tree->path);
  utils_tree->path = gtk_tree_path_to_string(path);
}

void
a_GtkUtils_tree_get_selected_data(XePGtkUtilsTree *utils_tree, 
				  const char **description,
				  int *type,
				  int *weight,
				  void **user_data1,
				  void **user_data2)
{
  if (description)
    *description = utils_tree->description;

  if (type)
    *type = utils_tree->type;

  if (weight)
    *weight   = utils_tree->weight;

  if (user_data1)
    *user_data1 = utils_tree->user_data1;

  if (user_data2)
    *user_data2 = utils_tree->user_data2;
}
				  
void
a_GtkUtils_tree_refresh(XePGtkUtilsTree *utils_tree)
{
  clistiter *list_iter;
  GtkTreeIter iter;
  XePGtkUtilsTreeItem *item;
  GtkTreePath *path;

  a_GtkUtils_tree_clear(utils_tree);

  if (utils_tree->refresh.callback)
    {
      utils_tree->refresh.callback(utils_tree, utils_tree->refresh.data);
    }
  else if (utils_tree->list)
    {
      for (list_iter = clist_begin(utils_tree->list);
	   list_iter; 
	   list_iter = list_iter->next)
	{
	  item = (XePGtkUtilsTreeItem *) list_iter->data;
	  gtk_tree_store_append(utils_tree->treestore, &iter, NULL);
	  gtk_tree_store_set(utils_tree->treestore, &iter, 
			     TREE_DESCRIPTION_COL, item->id,
			     TREE_TYPE_COL, 0,
			     TREE_WEIGHT_COL, PANGO_WEIGHT_NORMAL,
			     TREE_POINTER_COL, list_iter,
			     TREE_POINTER2_COL, NULL,
			     -1);
	}
    }
  else if (utils_tree->tree)
    {
      /* Root of the tree is not shown */
      GtkUtils_tree_append(utils_tree->treestore, NULL, utils_tree->tree);
    }

  path = gtk_tree_path_new_from_string(utils_tree->path ? 
				       utils_tree->path : "0");
  gtk_tree_view_set_cursor(GTK_TREE_VIEW(utils_tree->treeview), path, 
			   NULL, FALSE);
  gtk_tree_path_free(path);
}

static void
GtkUtils_tree_append(GtkTreeStore *treestore, 
		     GtkTreeIter *parent, 
		     cntree *ntree)
{
  GtkTreeIter child;
  cntree *tree_child;
  clistiter *list_iter;
  XePGtkUtilsTreeItem *item;

  if (ntree->children)
    {
      for (list_iter = clist_begin(ntree->children); 
	   list_iter;
	   list_iter = list_iter->next)
	{
	  tree_child = (cntree *) list_iter->data;
	  item = (XePGtkUtilsTreeItem *) ntree->data;
	  gtk_tree_store_append(treestore, &child, parent);
	  gtk_tree_store_set(treestore, &child, 
			     TREE_DESCRIPTION_COL, item->id,
			     TREE_TYPE_COL, 0,
			     TREE_WEIGHT_COL, PANGO_WEIGHT_NORMAL,
			     TREE_POINTER_COL, list_iter,
			     TREE_POINTER2_COL, NULL,
			     -1);
	  GtkUtils_tree_append(treestore, &child, tree_child);
	}
    }
}

void
a_GtkUtils_tree_clear(XePGtkUtilsTree *utils_tree)
{
  gtk_tree_store_clear(utils_tree->treestore);
}

void
a_GtkUtils_tree_set_add_callback(XePGtkUtilsTree *utils_tree,
				 XePGtkUtilsTreeCb callback,
				 void *data)
{
  GtkUtils_tree_set_callback(&utils_tree->add, callback, data);
}

void
a_GtkUtils_tree_set_remove_callback(XePGtkUtilsTree *utils_tree,
				    XePGtkUtilsTreeCb callback,
				    void *data)
{
  GtkUtils_tree_set_callback(&utils_tree->remove, callback, data);
}

void
a_GtkUtils_tree_set_edit_callback(XePGtkUtilsTree *utils_tree,
				  XePGtkUtilsTreeCb callback,
				  void *data)
{
  GtkUtils_tree_set_callback(&utils_tree->edit, callback, data);
}

void
a_GtkUtils_tree_set_top_callback(XePGtkUtilsTree *utils_tree,
				 XePGtkUtilsTreeCb callback,
				 void *data)
{
  GtkUtils_tree_set_callback(&utils_tree->top, callback, data);
}

void
a_GtkUtils_tree_set_changed_callback(XePGtkUtilsTree *utils_tree,
				     XePGtkUtilsTreeCb callback,
				     void *data)
{
  GtkUtils_tree_set_callback(&utils_tree->changed, callback, data);
}

void
a_GtkUtils_tree_set_refresh_callback(XePGtkUtilsTree *utils_tree,
				     XePGtkUtilsTreeCb callback,
				     void *data)
{
  GtkUtils_tree_set_callback(&utils_tree->refresh, callback, data);
}

void
a_GtkUtils_tree_set_expanded_callback(XePGtkUtilsTree *utils_tree,
				      XePGtkUtilsTreeCb callback,
				      void *data)
{
  GtkUtils_tree_set_callback(&utils_tree->expanded, callback, data);
}

static void
GtkUtils_tree_set_callback(XePGtkUtilsTreeButton *button,
			   XePGtkUtilsTreeCb callback,
			   void *data)
{
  button->callback = callback;
  button->data = data;
  if (button->button)
    gtk_widget_show(button->button);
}

static int
GtkUtils_tree_get_selected(XePGtkUtilsTree *utils_tree, 
			   GtkTreePath **path,
			   const char **description,
			   int *type,
			   int *weight,
			   void **user_data1,
			   void **user_data2)
{
  GtkTreeSelection *selection;
  GtkTreeModel *treemodel;
  GtkTreeIter iter;
  int r = -1;

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(utils_tree->treeview));

  if (gtk_tree_selection_get_selected(selection, &treemodel, &iter))
    {
      gtk_tree_model_get(treemodel, &iter, 
			 TREE_DESCRIPTION_COL, description,
			 TREE_TYPE_COL, type,
			 TREE_WEIGHT_COL, weight,
			 TREE_POINTER_COL, user_data1,
			 TREE_POINTER2_COL, user_data2,
			 -1);
      *path = gtk_tree_model_get_path(treemodel, &iter);
      r = 0;
    }
  return r;
}

/*! Create a new item and add it to the tree
 *  If the tree has a selected, the new item is inserted after it.
 */
static void 
GtkUtils_tree_add_cb(GtkButton *button, gpointer data)
{
  XePGtkUtilsTree *utils_tree = (XePGtkUtilsTree *) data;
  GtkTreePath *path;
  void *new_item;

  if (GtkUtils_tree_get_selected(utils_tree, &path, 
				 &utils_tree->description,
				 &utils_tree->type,
				 &utils_tree->weight,
				 &utils_tree->user_data1,
				 &utils_tree->user_data2) == 0)
    {
      gtk_tree_path_next(path);
    }
  else
    {
      utils_tree->description = NULL;
      utils_tree->type = 0;
      utils_tree->weight = 0;
      utils_tree->user_data1 = NULL;
      utils_tree->user_data2 = NULL;
      path = gtk_tree_path_new_first();
    }

  x_free(utils_tree->path);
  utils_tree->path = gtk_tree_path_to_string(path);
  gtk_tree_path_free(path);

  /* User callback to create the new item to insert 
   * If the callback returns NULL, do not add anything but
   * keep the insertion data available (path and list_iter) 
   * as the callback can be asynchronous. In which case
   * it is responsible to call a_GtkUtils_tree_add()
   * and it must make sure that nothing alters the
   * data of XePGtkUtilsTree
   */

  new_item = utils_tree->add.callback(utils_tree, utils_tree->add.data);

  if (new_item == NULL)
    {
      return;
    }
  a_GtkUtils_tree_finalize_add(utils_tree, new_item);
}

void 
a_GtkUtils_tree_finalize_add(XePGtkUtilsTree *utils_tree, void *new_item)
{
  if (utils_tree->list)
    {
      clistiter *iter = (clistiter *) utils_tree->user_data1;
     
      clist_insert_after(utils_tree->list, iter, new_item);
    }
  else if (utils_tree->tree)
    {
      clistiter *iter = (clistiter *) utils_tree->user_data1;
      cntree *parent = ((cntree *) (iter->data))->parent;
      cntree *child = cntree_new();
      child->data = new_item;
      child->parent = parent;
      clist_insert_after(parent->children, iter, child);
    }

  a_GtkUtils_tree_changed(utils_tree);
}

static void 
GtkUtils_tree_remove_cb(GtkButton *button, gpointer data)
{
  XePGtkUtilsTree *utils_tree = (XePGtkUtilsTree *) data;
  GtkTreePath *path = NULL;
  void *removed_item;

  if (GtkUtils_tree_get_selected(utils_tree, &path, 
				 &utils_tree->description,
				 &utils_tree->type,
				 &utils_tree->weight,
				 &utils_tree->user_data1,
				 &utils_tree->user_data2) != 0)
    {
      return;
    }

  /*  gtk_tree_path_prev(path);*/
  x_free(utils_tree->path);
  utils_tree->path = gtk_tree_path_to_string(path);
  gtk_tree_path_free(path);

  removed_item = utils_tree->remove.callback(utils_tree,
					     utils_tree->remove.data);

  /* NULL means removal succeeded, otherwise it is delayed or aborted */
   if (removed_item == NULL)
    {
      a_GtkUtils_tree_finalize_remove(utils_tree);
    }
}

void
a_GtkUtils_tree_finalize_remove(XePGtkUtilsTree *utils_tree)
{
  if (utils_tree->list)
    {
      clistiter *iter = (clistiter *) utils_tree->user_data1;
      clist_delete(utils_tree->list, iter);
    }
  else if (utils_tree->tree)
    {
      clistiter *iter = (clistiter *) utils_tree->user_data1;
      cntree *child = (cntree *) iter->data;
      cntree *parent = child->parent;
      if (cntree_remove_child(parent, child))
	{
	  A_LOGGING_SYSTEM_ERROR("child not present. Item removal ignored.");
	}
      else
	{
	  /* User callback is responsible to remove the children 
	   * of the item, and the child cntree must not have any child.
	   */
	  cntree_free(child, NULL, NULL);
	}
    }

  a_GtkUtils_tree_changed(utils_tree);
}

static void 
GtkUtils_tree_edit_cb(GtkButton *button, gpointer data)
{
  XePGtkUtilsTree *utils_tree = (XePGtkUtilsTree *) data;
  GtkTreePath *path = NULL;
  void *edited_item;

  if (GtkUtils_tree_get_selected(utils_tree, &path, 
				 &utils_tree->description,
				 &utils_tree->type,
				 &utils_tree->weight,
				 &utils_tree->user_data1,
				 &utils_tree->user_data2) != 0)
    {
      return;
    }

  x_free(utils_tree->path);
  utils_tree->path = gtk_tree_path_to_string(path);
  gtk_tree_path_free(path);

  edited_item = utils_tree->edit.callback(utils_tree, utils_tree->edit.data);

  if (edited_item)
    {
      a_GtkUtils_tree_finalize_edit(utils_tree, edited_item);
    }
}

void
a_GtkUtils_tree_finalize_edit(XePGtkUtilsTree *utils_tree, void *edited_item)
{
  if (utils_tree->list)
    {
      clistiter *iter = (clistiter *) utils_tree->user_data1;
     
      iter->data = edited_item;
    }
  else if (utils_tree->tree)
    {
      clistiter *iter = (clistiter *) utils_tree->user_data1;
      cntree *node = (cntree *) (iter->data);
      node->data = edited_item;
    }

  a_GtkUtils_tree_changed(utils_tree);
}

static void 
GtkUtils_tree_top_cb(GtkButton *button, gpointer data)
{
  XePGtkUtilsTree *utils_tree = (XePGtkUtilsTree *) data;
  GtkTreePath *path;
  void *top_item;
  
  if (GtkUtils_tree_get_selected(utils_tree, &path,
				 &utils_tree->description,
				 &utils_tree->type,
				 &utils_tree->weight,
				 &utils_tree->user_data1,
				 &utils_tree->user_data2) != 0)
    {
      return;
    }
  
  x_free(utils_tree->path);
  if (gtk_tree_path_up(path))
    {
      gtk_tree_path_down(path);
      utils_tree->path = gtk_tree_path_to_string(path);
    }
  else
    {
      utils_tree->path = g_strdup("0"); /* Root */
    }
  gtk_tree_path_free(path);

  if (utils_tree->top.callback)
    {
      top_item = utils_tree->top.callback(utils_tree, utils_tree->top.data);
    }
  else
    {
      top_item = ((clistiter *) (utils_tree->user_data1))->data;
    }
  
  if (top_item )
    {
      a_GtkUtils_tree_finalize_top(utils_tree);
    }
}
/*! Move the selected item at the first position of the same depth */
void
a_GtkUtils_tree_finalize_top(XePGtkUtilsTree *utils_tree)
{
  if (utils_tree->list)
    {
      clistiter *iter = (clistiter *) utils_tree->user_data1;
      void *item;
      item = iter->data;
      clist_delete(utils_tree->list, iter);
      clist_prepend(utils_tree->list, item);
    }
  else if (utils_tree->tree)
    {
      clistiter *iter = (clistiter *) utils_tree->user_data1;
      cntree *item;
      
      item = (cntree *) iter->data;
      cntree_remove_child(item->parent, item);
      cntree_prepend_child(item->parent, item);
    }

  a_GtkUtils_tree_changed(utils_tree);
}


static void 
GtkUtils_tree_collapsed_cb(GtkTreeView *treeview,GtkTreeIter *expanded_iter,
			   GtkTreePath *expanded_path, gpointer data)
{
  XePGtkUtilsTree *utils_tree = (XePGtkUtilsTree *) data;
  GtkTreeModel *treemodel = GTK_TREE_MODEL(utils_tree->treestore);
  if (utils_tree->expanded.callback == NULL)
    return;
  
  gtk_tree_model_get(treemodel, expanded_iter, 
		     TREE_DESCRIPTION_COL, &utils_tree->description,
		     TREE_TYPE_COL,        &utils_tree->type,
		     TREE_WEIGHT_COL,      &utils_tree->weight,
		     TREE_POINTER_COL,     &utils_tree->user_data1,
		     TREE_POINTER2_COL,    &utils_tree->user_data2,
		     -1);

  utils_tree->is_expanded = 0;

  utils_tree->expanded.callback(utils_tree, utils_tree->expanded.data);
}

static void 
GtkUtils_tree_expanded_cb(GtkTreeView *treeview,GtkTreeIter *expanded_iter,
			  GtkTreePath *expanded_path, gpointer data)
{
  XePGtkUtilsTree *utils_tree = (XePGtkUtilsTree *) data;
  GtkTreeModel *treemodel = GTK_TREE_MODEL(utils_tree->treestore);

  if (utils_tree->expanded.callback == NULL)
    return;

  /* TODO: check if this returns a copy of a string */
  gtk_tree_model_get(treemodel, expanded_iter, 
		     TREE_DESCRIPTION_COL, &utils_tree->description,
		     TREE_TYPE_COL,        &utils_tree->type,
		     TREE_WEIGHT_COL,      &utils_tree->weight,
		     TREE_POINTER_COL,     &utils_tree->user_data1,
		     TREE_POINTER2_COL,    &utils_tree->user_data2,
		     -1);

  utils_tree->is_expanded = 1;

  utils_tree->expanded.callback(utils_tree, utils_tree->expanded.data);
}

void
a_GtkUtils_tree_changed(XePGtkUtilsTree *utils_tree)
{
  a_GtkUtils_tree_refresh(utils_tree);
  
  if (utils_tree->changed.callback)
    {
      utils_tree->changed.callback(utils_tree, utils_tree->changed.data);
    }
}


GtkWidget *
a_GtkUtils_table_add_entry(GtkTable *table, const char *name, GtkWidget *label,
			   const char *help, guint col, guint row)
{
  GtkWidget *alabel;
  GtkWidget *entry;

  alabel = (name ? gtk_label_new_with_mnemonic(name) : label);

  gtk_misc_set_alignment(GTK_MISC(alabel), 0.0, 0.5);
  gtk_table_attach(table, alabel, col, col + 1, row, row + 1,
		   GTK_FILL, GTK_SHRINK, 0, 0);
  gtk_widget_show(alabel);
  
  entry = gtk_entry_new();
  gtk_label_set_mnemonic_widget(GTK_LABEL(alabel), entry);
  gtk_table_attach(table, entry, col + 1, col + 2, row, row + 1, 
		   GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(entry);
  
  if (help)
    { 
      a_GtkUtils_set_tip(entry, help);
    }

  return entry;
}

GtkWidget *
a_GtkUtils_table_add_entry_full(GtkTable *table, const char *name, 
				GtkWidget *label,
				const char *help, guint col, guint width,
				guint row, guint height)
{
  GtkWidget *alabel;
  GtkWidget *entry;

  alabel = (name ? gtk_label_new_with_mnemonic(name) : label);

  gtk_misc_set_alignment(GTK_MISC(alabel), 0.0, 0.5);
  gtk_table_attach(table, alabel, col, col + 1, row, row + height,
		   GTK_FILL, GTK_SHRINK, 0, 0);
  gtk_widget_show(alabel);

  entry = gtk_entry_new();
  gtk_label_set_mnemonic_widget(GTK_LABEL(alabel), entry);
  gtk_table_attach(table, entry, col + 1, col + 1 + width, row, row + height, 
		   GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(entry);
  
  if (help)
    {
      a_GtkUtils_set_tip(entry, help);
    }

  return entry;
}

void
a_GtkUtils_set_tip(GtkWidget *widget, const gchar *help)
{
  GtkTooltips *tips = a_GtkUtils_get_tips();

  gtk_tooltips_set_tip(tips, widget, help, NULL);
}

GtkTooltips *
a_GtkUtils_get_tips(void)
{
  if (tips)
    return tips;
  
  tips = gtk_tooltips_new();
  return tips;
}

#endif
