/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef LOGGING_H
#define LOGGING_H

#include <stdio.h>
#include <time.h>
#include <sys/times.h> /* for A_LOGGING_TIME */
#include "xetpan-types.h"

/*! Redirect standard error to a log file.
 *  This is done because many GUI application launchers do not
 *  show the user the standard error even if the application 
 *  did not start.
 *  Makes a backup of the old log file if it existed.
 *  \param path a directory where the log file will be created.
 *  \return 0 on success, system's <tt>errno</tt> otherwise.
 */
int
a_Logging_redirect(XetPan *xetpan, const char *path);

/*! Verbosity level. Currently: 0 to be quiet, 1, to be verbose.
 */
extern int verbose;

/*! Log filename. */
#define XETPAN_LOG_FILE "xetpan.log"

/*! Log's backup filename */
#define XETPAN_LOG_FILE_BAK "xetpan.log~"

#define SIZE_OF_TIME_STRING 9

/*! Error logging macro. */
#define A_LOGGING_ERROR(...) \
({char ts[SIZE_OF_TIME_STRING]; \
time_t t; \
t = time(NULL); \
strftime(ts, SIZE_OF_TIME_STRING, "%H:%M:%S", localtime(&t)); \
fprintf(stderr, "[%s] ERROR: ", ts); \
fprintf(stderr, __VA_ARGS__); \
fprintf(stderr, "\n");})

#define A_LOGGING_SYSTEM_ERROR(...) \
({fprintf(stderr, "%s (%s +%d) SYSTEM ERROR: ", __FUNCTION__, __FILE__, __LINE__), \
fprintf(stderr, __VA_ARGS__), \
fprintf(stderr, "\n");})

#define A_LOGGING_INFO(...) \
({if (verbose) \
{ \
char ts[SIZE_OF_TIME_STRING]; \
time_t t; \
t = time(NULL); \
strftime(ts, SIZE_OF_TIME_STRING, "%H:%M:%S", localtime(&t)); \
fprintf(stderr, "[%s] ", ts); \
fprintf(stderr, __VA_ARGS__); \
fprintf(stderr, "\n");} \
} )

/*! Debugging  macro. Used when verbosity is > 1 . */
#define A_LOGGING_DEBUG(...) \
({if (verbose > 1)  \
fprintf(stderr, "%s:%d %s: ",__FILE__ , __LINE__ , __FUNCTION__), \
fprintf(stderr, __VA_ARGS__), \
fprintf(stderr, "\n");})

#define A_LOGGING_TRACE \
({if (verbose > 1) \
fprintf(stderr,"%s:%d %s \n", __FILE__, __LINE__ , __FUNCTION__);})

/*! Timing macro. */
#define A_LOGGING_TIME(MSG)			\
  ({struct tms buf; times(&buf); printf("%.10f %.10f %s\n", ((float)buf.tms_utime)/sysconf(_SC_CLK_TCK), ((float)buf.tms_stime)/sysconf(_SC_CLK_TCK), MSG);})

#endif
