/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef XETPAN_H 
#define XETPAN_H

#include "xetpan-types.h"
#include "conf.h"
#include "ui-types.h"

#define UNIX_PATH_MAX 108

/*! Maximum size of a remote command string. */
#define REMOTE_MSG_SIZE 4096

/*! XetPan's main data structure. */
struct XetPan_s {
  char *config_path;                          /**< Configuration directory. */
  XePConf *config;                            /**< Configuration structure. */
  XePUI *ui;                                  /**< UI interfaces            */
  char lock_socket[UNIX_PATH_MAX];            /**< UNIX socket for remote 
					       *   control.                 */
  unsigned int server_watch_id;
  int server_fd;
  int log_fd;
  XePManager manager;
};

/*! Remote control information. */
struct XePRemote_s
{
  int fd;                  /**< Socket file descriptor. */
  unsigned int watch_id;   /**< Signal handler used by the GUI. */
  char *msg;
  XetPan *xetpan;
  ssize_t msg_size;
  ssize_t msg_len;
  ssize_t msg_idx;
};

extern XetPan *global_xetpan;

/*! Execute a remote command.
 *  Called when a remote command is received.
 *  \param xetpan the main XetPan structure.
 *  \param cmd the command identifier (name).
 *  \param option the argument of the option.
 *  \return XETPAN_NO_ERR if successful, XETPAN_ERR otherwise.
 */
int
a_XetPan_execute_remote_cmd(XePRemote *remote, char cmd, char *option);

/*! Check if XetPan should exits.
 *  XetPan exits when no more UI interfaces are running.
 */
int
a_XetPan_should_end(XetPan *xetpan);

/*! Close all windows and exits UI loop */
void
a_XetPan_quit(XetPan *xetpan);

#endif
